from magma import _BitType, BitType, BitsType, UIntType, SIntType


class MantleImportError(RuntimeError):
    pass


class UndefinedOperatorError(RuntimeError):
    pass


def raise_mantle_import_error_unary(self):
    raise MantleImportError(
        "Operators are not defined until mantle has been imported")


def raise_mantle_import_error_binary(self, other):
    raise MantleImportError(
        "Operators are not defined until mantle has been imported")


def define_raise_undefined_operator_error(type_str, operator, type_):
    if type_ == "unary":
        def wrapped(self):
            raise UndefinedOperatorError(
                f"{operator} is undefined for {type_str}")
    else:
        assert type_ == "binary"

        def wrapped(self, other):
            raise UndefinedOperatorError(
                f"{operator} is undefined for {type_str}")
    return wrapped


for op in ("__eq__", "__ne__"):
    setattr(_BitType, op, raise_mantle_import_error_binary)

for op in (
           "__and__",
           "__or__",
           "__xor__",
           "__invert__",
           "__add__",
           "__sub__",
           "__mul__",
           "__div__",
           "__lt__",
           # __le__ skipped because it's used for assignment on inputs
           # "__le__", 
           "__gt__",
           "__ge__"
           ):
    if op == "__invert__":
        setattr(_BitType, op,
                define_raise_undefined_operator_error("_BitType", op, "unary"))
    else:
        setattr(
            _BitType, op,
            define_raise_undefined_operator_error("_BitType", op, "binary"))


for op in ("__and__",
           "__or__",
           "__xor__",
           "__invert__"
           ):
    if op == "__invert__":
        setattr(BitType, op, raise_mantle_import_error_unary)
    else:
        setattr(BitType, op, raise_mantle_import_error_binary)


for op in ("__and__",
           "__or__",
           "__xor__",
           "__invert__",
           "__lshift__",
           "__rshift__",
           ):
    if op == "__invert__":
        setattr(BitsType, op, raise_mantle_import_error_unary)
    else:
        setattr(BitsType, op, raise_mantle_import_error_binary)

for op in ("__add__",
           "__sub__",
           "__mul__",
           "__div__",
           "__lt__",
           # __le__ skipped because it's used for assignment on inputs
           # "__le__",
           "__gt__",
           "__ge__"
           ):
    setattr(BitsType, op,
            define_raise_undefined_operator_error("BitsType", op, "binary"))

for op in ("__add__",
           "__sub__",
           "__mul__",
           "__div__",
           "__lt__",
           # __le__ skipped because it's used for assignment on inputs
           # "__le__",
           "__gt__",
           "__ge__"
           ):
    setattr(SIntType, op, raise_mantle_import_error_binary)
    setattr(UIntType, op, raise_mantle_import_error_binary)



import magma as m
from enum import Enum as pyEnum



class EnumMeta(type):
    def __new__(mcs, name, bases, attrs, **kwargs):
        cls = super().__new__(mcs, name, bases, attrs, **kwargs)
        kwargs = {}
        for key, value in attrs.items():
            if key[:2] == "__":
                continue
            kwargs[key] = value
        if not kwargs:
            return cls
        max_value = max(value for value in kwargs.values())
        num_bits = max(max_value.bit_length(), 1)
        type_ = m.Bits[num_bits]
        # TODO: Make Enum a subtype of Bits instead
        type_._is_magma_enum = True
        for key, value in kwargs.items():
            setattr(type_, key, m.bits(value, num_bits))
        return type_

class Enum(metaclass=EnumMeta):
    pass


# def Enum(**kwargs):
#     max_value = max(value for value in kwargs.values())
#     num_bits = max(max_value.bit_length(), 1)
#     type_ = m.Bits[num_bits]
#     for key, value in kwargs.items():
#         setattr(type_, key, m.bits(value, num_bits))
#     return type_



import functools
from collections.abc import Sequence
from .compatibility import IntegerTypes
from .t import In, Out, InOut, INPUT, OUTPUT, INOUT
from .bit import _BitKind, _BitType, Bit, BitType, VCC, GND
from .clock import ClockType, Clock, \
    Reset, ResetType, \
    AsyncReset, AsyncResetType, \
    Enable, EnableType
from .array import ArrayType, Array, ArrayKind
from .bits import BitsType, Bits, UIntType, UInt, SIntType, SInt, UIntKind, \
    SIntKind, BFloat
from .tuple import TupleType, tuple_ as tuple_imported, TupleKind, namedtuple
from .bitutils import int2seq
import magma as m
import hwtypes

__all__ = ['bit']
__all__ += ['clock', 'reset', 'enable', 'asyncreset']

__all__ += ['array']
__all__ += ['bits', 'uint', 'sint']

__all__ += ['tuple_', 'namedtuple']

__all__ += ['concat', 'repeat']
__all__ += ['sext', 'zext']


def can_convert_to_bit(value):
    return isinstance(value, (_BitType, ArrayType, TupleType, IntegerTypes))


def can_convert_to_bit_type(value):
    return isinstance(value, (_BitKind, ArrayKind, TupleKind))


def convertbit(value, totype, T):
    if isinstance(value, totype):
        return value

    if not can_convert_to_bit(value):
        raise ValueError(
            "bit can only be used on a Bit, an Array, or an int"
            f"; not {type(value)}")

    if isinstance(value, (ArrayType, TupleType)):
        if len(value) != 1:
            raise ValueError(
                "bit can only be used on arrays and tuples of length 1"
                f"; not {type(value)}")
        value = value[0]
        if not isinstance(value, _BitType):
            raise ValueError(
                "bit can only be used on arrays and tuples of bits"
                f"; not {type(value)}")

    assert isinstance(value, (IntegerTypes, _BitType))

    if isinstance(value, IntegerTypes):
        value = VCC if value else GND

    if value.isinput():
        b = In(T)(name=value.name)
    elif value.isoutput():
        b = Out(T)(name=value.name)
    else:
        b = T()
    b.port = value.port
    return b


def bit(value):
    return convertbit(value, BitType, Bit)


def clock(value):
    return convertbit(value, ClockType, Clock)


def reset(value):
    return convertbit(value, ResetType, Reset)


def asyncreset(value):
    return convertbit(value, AsyncResetType, AsyncReset)


def enable(value):
    return convertbit(value, EnableType, Enable)


def convertbits(value, n, totype, totypeconstructor, checkbit):
    if isinstance(value, totype):
        if n is not None and n != len(value):
            raise ValueError("converting a value should not change the size, use concat, zext, or sext instead.")
        return value

    convertible_types = (_BitType, TupleType, ArrayType, IntegerTypes,
                         Sequence)
    if not isinstance(value, convertible_types):
        raise ValueError(
            "bits can only be used on a Bit, an Array, a Tuple, an int, or a"
            f" Sequence; not : {type(value)}"
        )

    if isinstance(value, IntegerTypes):
        if n is None:
            n = max(value.bit_length(), 1)
        ts = int2seq(value, n)
    elif isinstance(value, Sequence):
        ts = list(value)
    elif isinstance(value, _BitType):
        if n is None:
            ts = [value]
        else:
            ts = n*[value]
    else:
        ts = [value[i] for i in range(len(value))]

    # create list of types
    Ts = []
    for t in ts:
        T = type(t)
        if T in IntegerTypes:
            T = Out(Bit)
        Ts.append(T)

    convert_to_bit = False
    # check that they are all the same
    for t in Ts:
        if checkbit:
            if not isinstance(t, _BitKind):
                raise ValueError(
                    "bits can only be used on Arrays or Tuples containing bits"
                    f", not : {t}")
        if t != T:
            if can_convert_to_bit_type(t) and can_convert_to_bit_type(T):
                # If they can all be converted to Bit, then that's okay
                convert_to_bit = True
            else:
                raise ValueError(
                    "All fields in a Array or a Tuple must be the same type"
                    f"got {t} expected {T}")

    assert len(Ts)

    if convert_to_bit is True:
        ts = [bit(t) for t in ts]
        T = {
            INPUT: In,
            OUTPUT: Out,
            INOUT: InOut
        }[T.direction](Bit)

    return totypeconstructor[len(Ts), T](*ts)


def array(value, n=None):
    return convertbits(value, n, ArrayType, Array, False)


def bits(value, n=None):
    return convertbits(value, n, BitsType, Bits, True)


def uint(value, n=None):
    if isinstance(value, SIntType):
        raise ValueError("uint cannot convert SInt")
    return convertbits(value, n, UIntType, UInt, True)


def sint(value, n=None):
    if isinstance(value, UIntType):
        raise ValueError("uint cannot convert SInt")
    return convertbits(value, n, SIntType, SInt, True)


def bfloat(value, n=None):
    return convertbits(value, n, BFloat, BFloat, True)


def concat(*arrays):
    ts = []
    for a in arrays:
        if isinstance(a, hwtypes.BitVector):
            ts.extend(a.bits())
        else:
            ts.extend(a.ts)
    return array(ts)


def repeat(value, n):
    if isinstance(value, BitType):
        repeats = bits(n * [value])
    else:
        repeats = array(n * [value])
    return repeats


def check_value_is_not_input(fn):
    @functools.wraps(fn)
    def wrapped(value, n):
        if isinstance(value, m.Type) and not value.isoutput():
            raise Exception(f"{fn.__name__} only works with non input values")
        return fn(value, n)
    return wrapped


# @check_value_is_not_input
def zext(value, n):
    assert isinstance(value, (UIntType, SIntType, BitsType)) or \
        isinstance(value, ArrayType) and isinstance(value.T, _BitKind)
    if isinstance(value, UIntType):
        zeros = uint(0, n)
    elif isinstance(value, SIntType):
        zeros = sint(0, n)
    elif isinstance(value, BitsType):
        zeros = bits(0, n)
    elif isinstance(value, ArrayType):
        zeros = array(0, n)
    result = concat(value, zeros)
    if isinstance(value, UIntType):
        return uint(result)
    elif isinstance(value, SIntType):
        return sint(result)
    elif isinstance(value, BitsType):
        return bits(result)
    return result


# @check_value_is_not_input
def sext(value, n):
    assert isinstance(value, SIntType)
    return sint(concat(array(value), array(value[-1], n)))


def tuple_(value, n=None):
    return tuple_imported(value, n)



from abc import abstractmethod
from .logging import error
from .port import Port, INPUT, OUTPUT, INOUT
from .t import Type, Kind
from .compatibility import IntegerTypes
from .debug import debug_wire, get_callee_frame_info
from .port import report_wiring_error

__all__ = ['_BitType', '_BitKind']
__all__ += ['BitType', 'BitKind']
__all__ += ['Bit', 'BitIn', 'BitOut', 'BitInOut']
__all__ += ['VCC', 'GND', 'HIGH', 'LOW']


class _BitType(Type):
    """
    Each bit is associated with a Port. The Port keeps track of how bits are
    wired together.
    """

    def __init__(self, *largs, **kwargs):
        super(_BitType, self).__init__(*largs, **kwargs)

        self.port = Port(self)

    __ne__ = Type.__ne__
    __hash__ = Type.__hash__

    def __call__(self, output):
        return self.wire(output, get_callee_frame_info())

    @classmethod
    def isoriented(cls, direction):
        return cls.direction == direction

    @debug_wire
    def wire(i, o, debug_info):
        # promote integer types to LOW/HIGH
        if isinstance(o, IntegerTypes):
            o = HIGH if o else LOW

        if not isinstance(o, _BitType):
            report_wiring_error(f'Cannot wire {i.debug_name} (type={type(i)}) to {o} (type={type(o)}) because {o.debug_name} is not a _Bit', debug_info)
            return

        i.port.wire(o.port, debug_info)
        i.debug_info = debug_info
        o.debug_info = debug_info

    def driven(self):
        return self.port.driven()

    def wired(self):
        return self.port.wired()

    # return the input or output _Bit connected to this _Bit
    def trace(self):
        t = self.port.trace()
        if t:
            t = t.bit
        return t

    # return the output _Bit connected to this input _Bit
    def value(self):
        t = self.port.value()
        if t:
            t = t.bit
        return t

    def const(self):
        return self is VCC or self is GND

    def getinst(self):
        t = self.trace()
        return t.name.inst if t else None

    def getgpio(self):
        return self.getinst()

    # return the bits driven by this bit
    def dependencies(self):
        assert self.isoutput(), "Trying to get dependencies from output bit"
        deps = map(lambda i: i.bit, self.port.wires.inputs)
        return deps

    def __repr__(self):
        if self is VCC:
            return '1'
        if self is GND:
            return '0'

        return self.name.qualifiedname(sep='.')

    def flatten(self):
        return [self]


class _BitKind(Kind):
    def __init__(cls, name, bases, dct):
        super(_BitKind, cls).__init__(name, bases, dct)

        if not hasattr(cls, 'direction'):
            cls.direction = None

    def __call__(cls, value=None, *args, **kwargs):
        if value is not None:
            if isinstance(value, (bool, IntegerTypes)):
                return VCC if value else GND
        result = super().__call__(*args, **kwargs)
        if value is not None:
            assert isinstance(value, BitType), type(value)
            result(value)
        return result

    def __eq__(cls, rhs):
        if not isinstance(rhs, _BitKind):
            return False

        return cls.direction == rhs.direction

    __ne__ = Kind.__ne__
    __hash__ = Kind.__hash__

    def __str__(cls):
        if cls.isinput():
            return 'In(_Bit)'
        if cls.isoutput():
            return 'Out(_Bit)'
        if cls.isinout():
            return 'InOut(_Bit)'
        return '_Bit'

    def size(self):
        return 1

    @abstractmethod
    def qualify(cls, direction):
        pass

    @abstractmethod
    def flip(cls):
        pass

    def get_family(cls):
        import magma as m
        return m.get_family()


class BitType(_BitType):
    __hash__ = Type.__hash__


class BitKind(_BitKind):
    def __eq__(cls, rhs):
        if not isinstance(rhs, BitKind):
            return False

        return cls.direction == rhs.direction

    __ne__ = _BitKind.__ne__
    __hash__ = _BitKind.__hash__

    def __str__(cls):
        if cls.isinput():
            return 'In(Bit)'
        if cls.isoutput():
            return 'Out(Bit)'
        if cls.isinout():
            return 'InOut(Bit)'
        return 'Bit'

    def qualify(cls, direction):
        if direction is None:
            return Bit
        elif direction == INPUT:
            return BitIn
        elif direction == OUTPUT:
            return BitOut
        elif direction == INOUT:
            return BitInOut
        return cls

    def flip(cls):
        if cls.isoriented(INPUT):
            return BitOut
        elif cls.isoriented(OUTPUT):
            return BitIn
        return cls


def MakeBit(**kwargs):
    return BitKind('Bit', (BitType,), kwargs)


Bit = MakeBit()
BitIn = MakeBit(direction=INPUT)
BitOut = MakeBit(direction=OUTPUT)
BitInOut = MakeBit(direction=INOUT)

VCC = BitOut(name="VCC")
GND = BitOut(name="GND")

HIGH = VCC
LOW = GND




from dataclasses import dataclass
from enum import Enum, auto
from .is_definition import isdefinition
from .logging import warning, error
from .passes import DefinitionPass


class MultipleDefinitionException(Exception):
    pass


class UniquificationMode(Enum):
    WARN = auto()
    ERROR = auto()
    UNIQUIFY = auto()


@dataclass(frozen=True)
class _HashStruct:
    defn_repr: str
    is_verilog: bool
    verilog_str: bool


def _make_hash_struct(definition):
    repr_ = repr(definition)
    if hasattr(definition, "verilogFile") and definition.verilogFile:
        return _HashStruct(repr_, True, definition.verilogFile)
    return _HashStruct(repr_, False, "")


def _hash(definition):
    hash_struct = _make_hash_struct(definition)
    return hash(hash_struct)


class UniquificationPass(DefinitionPass):
    def __init__(self, main, mode):
        super(UniquificationPass, self).__init__(main)
        self.definitions = {}
        self.mode = mode
        self.seen = {}
        self.original_names = {}

    def __call__(self, definition):
        name = definition.name
        key = _hash(definition)

        if name not in self.seen:
            self.seen[name] = {}
        if key not in self.seen[name]:
            if self.mode is UniquificationMode.UNIQUIFY and len(self.seen[name]) > 0:
                new_name = name + "_unq" + str(len(self.seen[name]))
                type(definition).rename(definition, new_name)
            self.seen[name][key] = [definition]
        else:
            if self.mode is not UniquificationMode.UNIQUIFY:
                assert self.seen[name][key][0].name == definition.name
            else:
                type(definition).rename(definition, self.seen[name][key][0].name)
            self.seen[name][key].append(definition)

    def _run(self, definition):
        if not isdefinition(definition):
            return

        for instance in definition.instances:
            instancedefinition = type(instance)
            if isdefinition(instancedefinition):
                self._run( instancedefinition )

        id_ = id(definition)
        if id_ in self.definitions:
            return
        self.definitions[id_] = definition
        self(definition)

    def run(self):
        super(UniquificationPass, self).run()
        duplicated = []
        for name, definitions in self.seen.items():
            if len(definitions) > 1:
                duplicated.append((name, definitions))
        UniquificationPass.handle(duplicated, self.mode)

    @staticmethod
    def handle(duplicated, mode):
        if len(duplicated):
            msg = f"Multiple definitions: {[name for name, _ in duplicated]}"
            if mode is UniquificationMode.ERROR:
                error(msg)
                raise MultipleDefinitionException([name for name, _ in duplicated])
            elif mode is UniquificationMode.WARN:
                warning(msg)


# This pass runs uniquification according to @mode and returns a dictionary
# mapping any renamed circuits to their original names. If @mode is ERROR or
# WARN the returned dictionary should be empty.
def uniquification_pass(circuit, mode):
    pass_ = UniquificationPass(circuit, mode)
    pass_.run()
    return pass_.original_names



from itertools import chain
from collections import OrderedDict
from .conversions import array
from .ref import AnonRef, InstRef, DefnRef
from .t import Type, Kind
from .port import INPUT, OUTPUT, INOUT
from .clock import ClockType, ClockTypes
from .array import ArrayType
from .tuple import TupleType
from .compatibility import IntegerTypes, StringTypes


__all__  = ['DeclareInterface']
__all__ += ['Interface']
__all__ += ['InterfaceKind']


def _flatten(l):
    """
    Flat an iterable of iterables to list.
    """
    return list(chain(*l))


def parse(decl):
    """
    Parse argument declaration of the form:

        (name0, type0, name1, type1, ..., namen, typen)
    """
    if len(decl) % 2:
        raise ValueError(f"Expected even number of arguments, got {len(decl)}")

    names = decl[::2]
    ports = decl[1::2]
    # If name is empty, convert to the index.
    names = [name if name else i for i, name in enumerate(names)]
    # Check that all ports are given as instances of Kind or Type.
    if not all(isinstance(port, (Kind, Type)) for port in ports):
        raise ValueError(f"Expected kinds or types, got {ports}")

    return names, ports


class _Interface(Type):
    """
    Abstract Base Class for an Interface.
    """
    def __str__(self):
        return str(type(self))

    def __repr__(self):
        s = ""
        for name, input in self.ports.items():
            if not input.isinput():
                continue
            output = input.value()
            if isinstance(output, (ArrayType, TupleType)):
                if not output.iswhole(output.ts):
                    for i in range(len(input)):
                        iname = repr(input[i])
                        oname = repr(output[i])
                        s += f"wire({oname}, {iname})\n"
                    continue
            iname = repr(input)
            oname = repr(output)
            s += f"wire({oname}, {iname})\n"
        return s

    @classmethod
    def items(cls):
        return cls.ports.items()

    def __iter__(self):
        return iter(self.ports)

    def __len__(self):
        return len(self.ports.keys())

    def __getitem__(self, key):
        if isinstance(key, str):
            return self.ports[key]
        if isinstance(key, int):
            return self.arguments()[key]
        if isinstance(key, slice):
            return array([self[i] for i in range(*key.indices(len(self)))])
        raise ValueError(f"Expected key as str, int, or slice, got {key} "
                         f"({type(key)})")

    def arguments(self):
        """Return all the argument ports."""
        return list(self.ports.values())

    def inputs(self, include_clocks=False):
        """Return all the argument input ports."""
        fn = lambda port: port.isinput() and \
            (not isinstance(port, ClockTypes) or include_clocks)
        return list(filter(fn, self.ports.values()))

    def outputs(self):
        """Return all the argument output ports."""
        return list(filter(lambda port: port.isoutput(), self.ports.values()))

    def args(self):
        """Return all the arguments as name, port."""
        return _flatten(self.ports.items())

    def decl(self):
        """
        Return all the arguments as name, flip(port) (same as the declaration).
        """
        return _flatten([name, type(port).flip()]
                        for name, port in self.ports.items())

    def inputargs(self):
        """Return all the input arguments as name, port."""
        return _flatten([name, port] for name, port in self.ports.items()
                        if port.isinput() and not isinstance(port, ClockTypes))

    def outputargs(self):
        """Return all the output arguments as name, port."""
        return _flatten([name, port] for name, port in self.ports.items()
                        if port.isoutput())

    def clockargs(self):
        """Return all the clock arguments as name, port."""
        return _flatten([name, port] for name, port in self.ports.items()
                        if isinstance(port, ClockTypes))

    def clockargnames(self):
        """Return all the clock argument names."""
        return [name for name, port in self.ports.items()
                if isinstance(port, ClockTypes)]

    def isclocked(self):
        """Return True if this interface has a Clock."""
        return any(isinstance(port, ClockType) for
                   port in self.ports.values())

class Interface(_Interface):
    """Interface class."""
    def __init__(self, decl, renamed_ports={}):
        """
        This function assumes the port instances are provided:
            e.g. Interface('I0', In(Bit)(), 'I1', In(Bit)(), 'O', Out(Bit)())
        """
        names, ports = parse(decl)
        args = OrderedDict()
        for name, port in zip(names, ports):
            if isinstance(name, IntegerTypes):
                name = str(name)  # convert integer to str, e.g. 0 to "0"
            if name in renamed_ports:
                raise NotImplementedError("Port renaming not implemented")
            args[name] = port
        self.ports = args

    def __str__(self):
        s = ", ".join(f"{k}: {v}" for k, v in self.ports.items())
        return f"Interface({s})"


class _DeclareInterface(_Interface):
    """
    _DeclareInterface class.

    First, an Interface is declared:
        Interface = DeclareInterface('I0', In(Bit), 'I1', In(Bit), 'O', Out(Bit))

    Then, the interface is instanced:
        interface = Interface()
    """
    def __init__(self, renamed_ports={}, inst=None, defn=None):
        # Parse the class Interface declaration.
        names, ports = parse(self.Decl)

        args = OrderedDict()
        for name, port in zip(names, ports):
            if   inst: ref = InstRef(inst, name)
            elif defn: ref = DefnRef(defn, name)
            else:      ref = AnonRef(name)

            if name in renamed_ports:
                ref.name = renamed_ports[name]
            if defn:
               port = port.flip()
            args[name] = port(name=ref)

        self.ports = args


class InterfaceKind(Kind):
    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        ports = []
        key = None
        for i, arg in enumerate(cls.Decl):
            if i % 2 == 0:
                key = arg
            else:
                ports.append((key, arg))
        cls.ports = OrderedDict(ports)

    def items(cls):
        return cls.ports.items()

    def __iter__(cls):
        return iter(cls.ports)

    def __str__(cls):
        args = []
        for i, arg in enumerate(cls.Decl):
            if i % 2 == 0:
                args.append(f"\"{arg}\"")
            else:
                args.append(str(arg))
        return ", ".join(args)

    def __eq__(cls, rhs):
        return cls.Decl == rhs.Decl

    __ne__=Kind.__ne__
    __hash__=Kind.__hash__


def DeclareInterface(*decl, **kwargs):
    """Interface factory function."""
    name = f"Interface({', '.join([str(a) for a in decl])})"
    dct = dict(Decl=decl, **kwargs)
    return InterfaceKind(name, (_DeclareInterface,), dct)



from collections.abc import Sequence, Mapping
from collections import OrderedDict
from .ref import AnonRef, TupleRef
from .t import Type, Kind
from .compatibility import IntegerTypes
from .bit import BitOut, VCC, GND
from .debug import debug_wire, get_callee_frame_info
from .port import report_wiring_error

__all__  = ['TupleType', 'TupleKind', 'Tuple']

#
# Create an Tuple
#
#  Tuple()
#  - creates a new tuple value
#  Tuple(v0, v1, ..., vn)
#  - creates a new tuple value where each field equals vi
#
class TupleType(Type):
    def __init__(self, *largs, **kwargs):

        Type.__init__(self, **kwargs) # name=

        self.ts = []
        if len(largs) > 0:
            assert len(largs) == self.N
            for i in range(self.N):
                k = str(self.Ks[i])
                t = largs[i]
                if isinstance(t, IntegerTypes):
                    t = VCC if t else GND
                assert type(t) == self.Ts[i]
                self.ts.append(t)
                setattr(self, k, t)
        else:
            for i in range(self.N):
                k = str(self.Ks[i])
                T = self.Ts[i]
                t = T(name=TupleRef(self,k))
                self.ts.append(t)
                setattr(self, k, t)

    __hash__ = Type.__hash__

    def __eq__(self, rhs):
        if not isinstance(rhs, TupleType): return False
        return self.ts == rhs.ts

    def __repr__(self):
        if not isinstance(self.name, AnonRef):
            return repr(self.name)
        ts = [repr(t) for t in self.ts]
        kts = ['{}={}'.format(k, v) for k, v in zip(self.Ks, ts)]
        return 'tuple(dict({})'.format(', '.join(kts))

    def __getitem__(self, key):
        if key in self.Ks:
            key = self.Ks.index(key)
        return self.ts[key]

    def __len__(self):
        return self.N

    def __call__(self, o):
        return self.wire(o, get_callee_frame_info())

    @classmethod
    def isoriented(cls, direction):
        for T in cls.Ts:
            if not T.isoriented(direction):
                return False
        return True


    @debug_wire
    def wire(i, o, debug_info):
        # print('Tuple.wire(', o, ', ', i, ')')

        if not isinstance(o, TupleType):
            report_wiring_error(f'Cannot wire {o.debug_name} (type={type(o)}) to {i.debug_name} (type={type(i)}) because {o.debug_name} is not a Tuple', debug_info)  # noqa
            return

        if i.Ks != o.Ks:
            report_wiring_error(f'Cannot wire {o.debug_name} (type={type(o)}, keys={i.Ks}) to {i.debug_name} (type={type(i)}, keys={o.Ks}) because the tuples do not have the same keys', debug_info)  # noqa
            return

        #if i.Ts != o.Ts:
        #    print('Wiring error: Tuple elements must have the same type')
        #    return

        for i_elem, o_elem in zip(i, o):
            if o_elem.isinput():
                o_elem.wire(i_elem, debug_info)
            else:
                i_elem.wire(o_elem, debug_info)

    def driven(self):
        for t in self.ts:
            if not t.driven():
                return False
        return True

    def wired(self):
        for t in self.ts:
            if not t.wired():
                return False
        return True

    # test whether the values refer a whole tuple
    def iswhole(self, ts):

        for i in range(len(ts)):
            if ts[i].anon():
                #print('not an inst or defn')
                return False

        for i in range(len(ts)):
            # elements must be an tuple reference
            if not isinstance(ts[i].name, TupleRef):
                #print('not an tuple ref')
                return False

        for i in range(1,len(ts)):
            # elements must refer to the same tuple
            if ts[i].name.tuple is not ts[i-1].name.tuple:
                return False

        for i in range(len(ts)):
            # elements should be numbered consecutively
            if ts[i].name.index != self.Ks[i]:
                return False

        if len(ts) != len(ts[0].name.tuple):
            # elements should refer to a whole tuple
            return False

        return True


    def trace(self):
        ts = [t.trace() for t in self.ts]

        for t in ts:
            if t is None:
                return None

        if len(ts) == self.N and self.iswhole(ts):
            return ts[0].name.tuple

        return tuple_(dict(zip(self.Ks,ts)))

    def value(self):
        ts = [t.value() for t in self.ts]

        for t in ts:
            if t is None:
                return None

        if len(ts) == self.N and self.iswhole(ts):
            return ts[0].name.tuple

        return tuple_(dict(zip(self.Ks,ts)))

    def flatten(self):
        return sum([t.flatten() for t in self.ts], [])

    def const(self):
        for t in self.ts:
            if not t.const():
                return False

        return True

    def keys(self):
        return self.Ks

    def values(self):
        return self.ts

    def items(self):
        return zip(self.Ks, self.ts)

class TupleKind(Kind):
    def __init__(cls, name, bases, dct):
        super(TupleKind, cls).__init__(name, bases, dct)
        cls.N = len(cls.Ks)

    def __str__(cls):
        args = []
        for i in range(cls.N):
             args.append("%s=%s" % (cls.Ks[i], str(cls.Ts[i])))
        return "Tuple(%s)" % ",".join(args)

    def __eq__(cls, rhs):
        if not isinstance(rhs, TupleKind): return False

        if cls.Ks != rhs.Ks: return False
        if cls.Ts != rhs.Ts: return False

        return True

    __ne__ = Kind.__ne__
    __hash__ = Kind.__hash__

    def __len__(cls):
        return cls.N

    def __getitem__(cls, key):
        if key in cls.Ks:
            key = cls.Ks.index(key)
        return cls.Ts[key]

    def size(cls):
        n = 0
        for T in cls.Ts:
            n += T.size()
        return n

    def qualify(cls, direction):
        if cls.isoriented(direction):
            return cls
        return Tuple(OrderedDict(zip(cls.Ks, [T.qualify(direction) for T in cls.Ts])))

    def flip(cls):
        return Tuple(OrderedDict(zip(cls.Ks, [T.flip() for T in cls.Ts])))


#
# Tuple(Mapping)
# Tuple(Sequence)
#
# *largs with largs being comprised of magma types
#    Tuple(T0, T1, ..., Tn)
#
# **kwargs - only called if largs is empty
#    Tuple(x=Bit, y=Bit) -> Tuple(**kwargs)
#
def Tuple(*largs, **kwargs):
    if largs:
        if isinstance(largs[0], Kind):
            Ks = range(len(largs))
            Ts = largs
        else:
            largs = largs[0]
            if isinstance(largs, Sequence):
                Ks = range(len(largs))
                Ts = largs
            elif isinstance(largs, Mapping):
                Ks = list(largs.keys())
                Ts = list(largs.values())
            else:
                assert False
    else:
        Ks = list(kwargs.keys())
        Ts = list(kwargs.values())

    # check types and promote integers
    for i in range(len(Ts)):
        T = Ts[i]
        if T in IntegerTypes:
            T = BitOut
            Ts[i] = T
        if not isinstance(T, Kind):
            raise ValueError(f'Tuples must contain magma types - got {T}')

    name = 'Tuple(%s)' % ", ".join([f"{k}: {t}" for k, t in zip(Ks, Ts)])
    return TupleKind(name, (TupleType,), dict(Ks=Ks, Ts=Ts))

from .bitutils import int2seq
from .array import ArrayType, Array
from .bit import _BitKind, _BitType, Bit, BitKind, BitType, VCC, GND

#
# convert value to a tuple
#   *value = tuple from positional arguments
#   **kwargs = tuple from keyword arguments
#
def tuple_(value, n=None):
    if isinstance(value, TupleType):
        return value

    if not isinstance(value, (_BitType, ArrayType, IntegerTypes, Sequence, Mapping)):
        raise ValueError(
            "bit can only be used on a Bit, an Array, or an int; not {}".format(type(value)))

    decl = OrderedDict()
    args = []

    if isinstance(value, IntegerTypes):
        if n is None:
            n = max(value.bit_length(),1)
        value = int2seq(value, n)
    elif isinstance(value, _BitType):
        value = [value]
    elif isinstance(value, ArrayType):
        value = [value[i] for i in range(len(value))]

    if isinstance(value, Sequence):
        ts = list(value)
        for i in range(len(ts)):
            args.append(ts[i])
            decl[i] = type(ts[i])
    elif isinstance(value, Mapping):
        for k, v in value.items():
            args.append(v)
            decl[k] = type(v)

    return Tuple(decl)(*args)


def namedtuple(**kwargs):
    return tuple_(kwargs)



import sys
from abc import abstractmethod
if sys.version_info < (3, 4):
    import abc
    ABC = abc.ABCMeta('ABC', (object,), {})
else:
    from abc import ABC
from collections import namedtuple
from itertools import product
from .simulator import CircuitSimulator, ExecutionState
from ..transforms import flatten, setup_clocks
from ..circuit import *
from ..scope import *
from ..bit import VCC, GND, BitType, _BitType
from ..array import ArrayType
from ..bits import SIntType, BitsType, UIntType
from hwtypes import BitVector
import hwtypes
from ..bitutils import seq2int
from ..clock import ClockType

__all__ = ['PythonSimulator']

ExecutionOrder = namedtuple('ExecutionOrder', ['stateful', 'combinational'])


class PythonSimulatorException(Exception):
    pass


class SimPrimitive:
    def __init__(self, primitive, value_store):
        if primitive.simulate is None:
            raise ValueError("Cannot simulate {} of type {} because it does not have a Python simulate method".format(primitive, type(primitive)))
        self.primitive = primitive
        self.inputs = []
        self.outputs = []
        self.value_store = value_store
        self.state_store = {}

        for bit in self.primitive.interface.ports.values():
            if not isinstance(bit, ArrayType):
                bit = [bit]
            for b in bit:
                if b.isinput():
                    self.inputs.append(b)
                else:
                    self.outputs.append(b)

    def stateful(self):
        return self.primitive.stateful

    def inputs_satisfied(self):
        for i in self.inputs:
            if not self.value_store.value_initialized(i):
                return False

        return True

    def initialize_outputs(self):
        # Initializes all outputs to False, should perhaps
        # initialize based on starting inputs?
        for o in self.outputs:
            self.value_store.set_value(o, False)

    def simulate(self):
        self.primitive.simulate(self.value_store, self.state_store)

class WatchPoint:
    idx = 0
    def __init__(self, bit, scope, simulator, value):
        self.bit = bit
        self.scope = scope
        self.simulator = simulator
        self.value = value
        self.old_val = self.simulator.get_value(self.bit, self.scope)

        WatchPoint.idx += 1
        self.idx = WatchPoint.idx

    def was_triggered(self):
        new_val = self.simulator.get_value(self.bit, self.scope)
        triggered = new_val != self.old_val
        if self.value:
            if self.value != new_val:
                triggered = False
        self.old_val = new_val

        return triggered

class ValueStore:
    def __init__(self):
        self.value_map = {}

    def value_initialized(self, bit):
        if isinstance(bit, ArrayType):
            for b in bit:
                if not self.value_initialized(b):
                    return False

            return True

        if bit.isinput():
            bit = bit.value()

        if bit.const():
            return True

        return bit in self.value_map

    def get_value(self, bit):
        if isinstance(bit, ArrayType):
            value = [self.get_value(b) for b in bit]
            if isinstance(bit, SIntType):
                return BitVector[len(bit)](value).as_sint()
            elif isinstance(bit, UIntType):
                return BitVector[len(bit)](value).as_uint()
            elif isinstance(bit, BitsType):
                return BitVector[len(bit)](value)
            return value

        if bit.isinput():
            bit = bit.value()

        if bit.const():
            return True if bit is VCC else False

        return self.value_map[bit]

    def set_value(self, bit, newval):
        if not bit.isoutput():
            raise TypeError("Can only call set value on an input")

        if isinstance(bit, ArrayType):
            if isinstance(newval, BitVector):
                newval = newval.as_bool_list()
            elif isinstance(newval, BitsType):
                if not newval.const():
                    raise ValueError("Calling set_value with a BitsType only works with a constant")
                newval = newval.bits()
            elif isinstance(bit, BitsType) and isinstance(newval, int):
                if not isinstance(bit, SIntType) and newval < 0:
                    raise ValueError(f"Can only set {bit} of type {type(bit)} with positive integer, not {newval}")
                newval = BitVector[len(bit)](newval).as_bool_list()
            elif not isinstance(newval, list):
                raise TypeError(f"Calling set_value with {bit} of type {type(bit)} only works with a list of values or a BitVector")

            for b,v in zip(bit, newval):
                self.set_value(b, v)
            return

        if isinstance(newval, int) and newval in {0, 1}:
            newval = bool(newval)
        if isinstance(newval, hwtypes.Bit):
            newval = bool(newval)
        if not isinstance(newval, bool):
            raise TypeError(f"Can only set Bit {bit} with a boolean value or 0 or 1, not {newval} (type={type(newval)})")

        self.value_map[bit] = newval

class PythonSimulator(CircuitSimulator):
    def __setup_primitives(self):
        wrapped = []
        for primitive in self.circuit.instances:
            wrapped.append(SimPrimitive(primitive, self.value_store))

        return wrapped

    def initialize(self, bit):
        if isinstance(bit, ArrayType):
            for b in bit:
                self.initialize(b)
        else:
            if bit.isoutput():
                self.circuit_inputs.append(bit)
                self.value_store.set_value(bit, False)
            else:
                self.circuit_outputs.append(bit)

    def __setup_circuit(self, clock):
        if clock is not None:
            clock = self.txfm.get_new_bit(clock, self.default_scope)
        self.clock = clock

        self.circuit_inputs = []
        self.circuit_outputs = []
        for name, bit in self.circuit.interface.ports.items():
            self.initialize(bit)

    def __outputs_initialized(self):
        for bit in self.circuit_outputs:
            assert bit.isinput()
            if not self.value_store.value_initialized(bit):
                return False

        return True

    def __sort_state_primitives(self, state_primitives):
        """
        State primitives should be sorted in reversed topological order.
        This ensures that the simulation order of the stateful elements
        is correct.
        Intuition:
            If the output value of a state element `x` feeds into the input of
            another state element `y`,
            `y` should perform it's simulation before `x` because it will use
            the value of the signal on the previous clock cycle.
        """
        sorted_state_primitives = []
        for primitive_1 in state_primitives:
            inserted = False
            for index, primitive_2 in enumerate(sorted_state_primitives):
                for output in primitive_1.inputs:
                    if any(output.value() is x for x in primitive_2.inputs):
                        sorted_state_primitives.insert(index, primitive_1)
                        inserted = True
                        break
                if inserted:
                    break
            if not inserted:
                sorted_state_primitives.append(primitive_1)
        sorted_state_primitives.reverse()
        return sorted_state_primitives

    def __get_ordered_primitives(self, unordered_primitives):
        state_primitives = []
        after_state = []

        state_primitives = []
        for primitive in unordered_primitives:
            if primitive.stateful():
                primitive.initialize_outputs()
                state_primitives.append(primitive)

        sorted_state_primitives = self.__sort_state_primitives(state_primitives)

        unordered_primitives[:] = [u for u in unordered_primitives if not u.stateful()]

        combinational = []
        while len(unordered_primitives) > 0:
            found = False
            for primitive in unordered_primitives:
                if primitive.inputs_satisfied():
                    primitive.initialize_outputs()
                    combinational.append(primitive)
                    unordered_primitives.remove(primitive)
                    found = True
                    break
            assert found, "Some circuits have unsatisfied inputs"

        return ExecutionOrder(stateful=sorted_state_primitives, combinational=combinational)

    def __step(self):
        if self.clock is None:
            raise PythonSimulatorException("Cannot step a simulated circuit "
                    "without a clock, did you pass a clock during "
                    "initialization?")
        cur_clock_val = self.value_store.get_value(self.clock)
        self.value_store.set_value(self.clock, not cur_clock_val)

    def __init__(self, main_circuit, clock=None):
        if isinstance(main_circuit, CircuitType):
            raise ValueError("PythonSimulator must be called with a Circuit definition, not an instance")
        if clock is not None and not isinstance(clock, ClockType):
            raise ValueError("clock must be a ClockType or None")
        setup_clocks(main_circuit)
        self.main_circuit = main_circuit
        self.txfm = flatten(main_circuit)
        self.circuit = self.txfm.circuit
        self.value_store = ValueStore()
        self.default_scope = Scope()
        self.__setup_circuit(clock)
        self.watchpoints = []

        primitives = self.__setup_primitives()

        self.execution_order = self.__get_ordered_primitives(primitives)

        assert self.__outputs_initialized(), "All circuit outputs not initialized."

    def get_capabilities(self):
        return []

    def get_value(self, bit, scope=None):
        if scope is None:
            scope = self.default_scope
        newbit = self.txfm.get_new_bit(bit, scope)
        if newbit is None:
            return None

        try:
            return self.value_store.get_value(newbit)
        except KeyError:
            return None

    def set_value(self, bit, newval, scope=None):
        if scope is None:
            scope = self.default_scope
        newbit = self.txfm.get_new_bit(bit, scope)
        if not self.is_circuit_input(newbit):
            message = "Only setting main's inputs is supported (Trying to set: {})".format(bit)
            raise PythonSimulatorException(message)
        else:
            self.value_store.set_value(newbit, newval)

    def is_circuit_input(self, value):
        """
        Checks if `value` is in `self.circuit_inputs`.
        If `value` is an `ArrayType`, it recursively checks the elements
        """
        if isinstance(value, _BitType):
            return any(value is x for x in self.circuit_inputs)
        elif isinstance(value, ArrayType):
            return all(self.is_circuit_input(elem) for elem in value)
        else:
            raise NotImplementedError(type(value))

    def advance(self, n=1):
        cycles = 0
        for i in range(0, n):
            self.__step()
            state = self.evaluate()

            if not state.clock:
                cycles += 1
            if state.triggered_points:
                return ExecutionState(triggered_points=state.triggered_points, clock=state.clock, cycles=cycles)

        return ExecutionState(triggered_points=[], clock=self.get_clock_value(), cycles=cycles)

    def evaluate(self, no_update=False):
        for primitive in self.execution_order.stateful:
            primitive.simulate()
        for primitive in self.execution_order.combinational:
            primitive.simulate()

        triggered = []
        for watch in self.watchpoints:
            if watch.was_triggered():
                triggered.append(watch)

        return ExecutionState(triggered_points=triggered, clock=self.get_clock_value(), cycles=0)

    def get_clock_value(self):
        """
        Looks up the value of `self.clock` in `self.value_store`
        Returns None if self.clock is None (circuit doesn't have a clock)
        """
        if self.clock is not None:
            return self.value_store.get_value(self.clock)
        return None

    def rewind(self, halfcycles):
        raise PythonSimulatorException("Reversing not currently supported")

    def cont(self):
        cycles = 0
        while True:
            self.__step()
            state = self.evaluate()
            if not state.clock:
                cycles += 1

            if state.triggered_points:
                return ExecutionState(triggered_points=state.triggered_points, clock=state.clock, cycles=cycles)

    def add_watchpoint(self, bit, scope, value=None):
        self.watchpoints.append(WatchPoint(bit, scope, self, value))
        return self.watchpoints[-1].idx

    def delete_watchpoint(self, num):
        for w in self.watchpoints:
            if w.idx == num:
                del w
                return True

        return False

    def __call__(self, *largs):
        circuit = self.main_circuit

        j = 0
        for name, port in circuit.interface.ports.items():
            if port.isoutput():
                val = largs[j]
                if isinstance(port, ArrayType):
                    n = type(port).N
                    val = BitVector[n](val)
                self.set_value(getattr(circuit, name), val)
                j += 1

        self.evaluate()

        outs = []
        for name, port in circuit.interface.ports.items():
            if port.isinput():
                val = self.get_value(getattr(circuit, name))
                val = seq2int(val) if isinstance(val, list) else int(val) 
                outs.append(val)

        if len(outs) == 1:
            return outs[0]
        return tuple(outs)



import os
import types
import operator
from functools import reduce
from collections import OrderedDict
from collections.abc import Sequence
from ..port import INPUT, OUTPUT, INOUT, flip
from ..ref import DefnRef
from ..compatibility import IntegerTypes
from ..bit import _BitType, _BitKind, VCC, GND
from ..clock import ClockType, EnableType, ResetType
from ..array import ArrayKind, ArrayType
from ..bits import SIntType
from ..tuple import TupleType
from ..is_definition import isdefinition
from ..clock import wiredefaultclock
import logging
import os

logger = logging.getLogger('magma').getChild('verilog_backend')
level = os.getenv("MAGMA_VERILOG_BACKEND_LOG_LEVEL", "WARN")
# TODO: Factor this with magma.logging code for debug level validation
if level in ["DEBUG", "WARN", "INFO"]:
    logger.setLevel(getattr(logging, level))
elif level is not None:
    logging.warning("Unsupported value for MAGMA_VERILOG_BACKEND_LOG_LEVEL:"
                    f" {level}")

from .util import get_codegen_debug_info, make_relative

#__all__  = ['hstr', 'bstr']
__all__  = ['hstr']

# return the hex character for int n
def hex(n):
    if n < 10: return chr(ord('0')+n)
    else:      return chr(ord('A')+n-10)

# return a hex string reprenting n
def hstr(n, bits):
    format = "%d'h" % bits
    nformat = []
    n &= (1 << bits)-1
    for i in range((bits+3)//4):
        nformat.append(n%16)
        n //= 16
    nformat.reverse()
    return format + reduce(operator.add, map(hex, nformat))

def bstr(n, bits):
    if bits == 1:
        return "1'b1" if init else "1'b0"
    format = "%d'b" % bits
    nformat = []
    n &= (1 << bits)-1
    for i in range(bits):
        nformat.append(n%2)
        n //= 2
    nformat.reverse()
    return format + reduce(operator.add, map(hex, nformat))

# return the verilog name of a data value
def vname(t):
    if t is VCC: return "1'b1"
    if t is GND: return "1'b0"

    if isinstance(t, ArrayType):
        # print(t.ts)
        if not t.iswhole(t.ts):
            # the sequence of values is concantenated
            t = [vname(i) for i in t.ts]
            t.reverse()
            return '{' + ','.join(t) + '}'

    assert not t.anon(), (t.name)
    return t.name.qualifiedname(sep='_')

# return the verilog declaration for the data type
def vdecl(t):
    if isinstance(t, ArrayType):
        signed = "signed " if isinstance(t, SIntType) else ""
        return '{}[{}:{}]'.format(signed, t.N-1, 0)
    else:
        assert isinstance(t, _BitType)
        return ""

# return the verilog module args
def vmoduleargs(self):
    def append(args, port, name):
        if   port.isinput():  d = OUTPUT
        elif port.isoutput(): d = INPUT
        else: d = INOUT
        args.append("%s %s %s" % (d, vdecl(port), name))

    args = []
    for name, port in self.ports.items():
        if isinstance(port, TupleType):
            for i in range(len(port)):
                append(args, port[i], vname(port[i]))
        else:
            #d = flip(port.direction)
            append(args, port, name)
    return args


def compileinstance(self):

    # print('compileinstance', str(self), str(type(self)))

    def arg(k,v):
        if not isinstance(v, str): v = str(v)
        return '.%s(%s)' % (k, v)

    args = []
    debug_str = ""
    for k, v in self.interface.ports.items():
        if getattr(v, "debug_info", False) and get_codegen_debug_info():
            filename, lineno, module = v.debug_info
        #print('arg', k, v,)
        if v.isinput():
            # find the output connected to v
            w = v.value()
            if not w:
                logging.warning(f'{v.debug_name} not connected')
                continue
            v = w
        if isinstance(v, TupleType):
            for i in range(len(v)):
                args.append(arg('%s_%s' %
                    (v[i].name.tuple.name, v[i].name.index), vname(v[i])))
        elif isinstance(k, IntegerTypes):
            args.append( vname(v) )
        else:
            args.append( arg(k,vname(v)) )
        if getattr(v, "debug_info", False) and get_codegen_debug_info():
            debug_str += f"// Argument {k}({vname(v)}) wired at {make_relative(filename)}:{lineno}\n"

    params = []
    for k, v in self.kwargs.items():
       if k not in {'loc', 'name', 'T'}:
           if isinstance(v, tuple):
               v = hstr(v[0], v[1])
           params.append(arg(k, v))
    params = sorted(params)

    #s = '(* loc="%d,%d/%d" *)\n' % self.loc if self.loc else ""

    s = str(self.__class__.verilog_name)
    if len(params):
        if len(params) > 2:
            s += ' #(' + ",\n".join(params) + ')'
        else:
            s += ' #(' + ", ".join(params) + ')'
    s += ' ' + str(self.name)

    return debug_str + '%s (%s)' % (s, ', '.join(args))

def compiledefinition(cls):

    # for now only allow Bit or Array(n, Bit)
    for name, port in cls.interface.ports.items():
        if isinstance(port, (ArrayKind, ArrayType)):
            if not isinstance(port.T, (_BitType, _BitKind)):
                raise Exception(f'Argument {cls.__name__}.{name} of type {type(port)} is not supported, the verilog backend only supports simple 1-d array of bits of the form Array(N, Bit)')


    if cls.verilogFile:
       s = cls.verilogFile
    else:
        args = ', '.join(vmoduleargs(cls.interface))
        s = ''
        if get_codegen_debug_info() and cls.debug_info.filename and cls.debug_info.lineno:
            s += f'// Defined at {make_relative(cls.debug_info.filename)}:{cls.debug_info.lineno}\n'
        s += 'module %s (%s);\n' % (cls.verilog_name, args)
        if cls.verilog:
            s += cls.verilog + '\n'
            if cls.verilogLib:
                import re
                for libName in cls.verilogLib:
                    if re.search("\.v$",libName):
                        with open(libName,'r') as libFile:
                            s = libFile.read() + s
                    else:
                        s = libName + s
        else:
            def wire(port):
                return 'wire %s %s;\n' % (vdecl(port), vname(port))

            # declare a wire for each instance output
            for instance in cls.instances:
                for port in instance.interface.ports.values():
                    if isinstance(port, TupleType):
                        for i in range(len(port)):
                            s += wire(port[i])
                    else:
                        if not port.isinput():
                            s += wire(port)

            #print('compile instances')
            # emit the structured verilog for each instance
            for instance in cls.instances:
                wiredefaultclock(cls, instance)
                if getattr(instance, "debug_info", False) and \
                        instance.debug_info.filename and instance.debug_info.lineno and \
                        get_codegen_debug_info():
                    s += f"// Instanced at {make_relative(instance.debug_info.filename)}:{instance.debug_info.lineno}\n"
                s += compileinstance(instance) + ";\n"

            # assign to module output arguments
            for port in cls.interface.ports.values():
                if port.isinput():
                    output = port.value()
                    if output:
                        if isinstance(output, TupleType):
                            for name, input in cls.interface.ports.items():
                                if input.isinput():
                                    output = input.value()
                                    assert isinstance(output, TupleType)
                                    for i in range(len(input)):
                                        iname = vname(input[i])
                                        oname = vname(output[i])
                                        s += 'assign %s = %s;\n' % (iname, oname)
                        else:
                            iname = vname(port)
                            oname = vname(output)
                            if getattr(port, "debug_info", False) and get_codegen_debug_info():
                                s += f"// Wired at {make_relative(port.debug_info[0])}:{port.debug_info[1]}\n"
                            s += 'assign %s = %s;\n' % (iname, oname)
                    else:
                        logging.warning(f"{cls.__name__}.{port.name} is unwired")

        s += "endmodule\n"

    return s

def find(circuit, defn):
    if not isdefinition(circuit):
        return defn
    for i in circuit.instances:
        find(type(i), defn)
    name = circuit.verilog_name
    if name not in defn:
        defn[name] = circuit
    return defn


def compile(main):
    defn = find(main,OrderedDict())

    code = ''

    for k, v in defn.items():
         logging.debug(f'compiling circuit {k}')
         code += compiledefinition(v) + '\n'
    return code



import types
from collections import OrderedDict
from collections.abc import Sequence
from ..bitutils import seq2int, fun2seq
from ..bit import VCC, GND, BitType
from ..array import ArrayType
from ..port import INPUT, OUTPUT, INOUT
from ..clock import wiredefaultclock
from ..is_definition import isdefinition

__all__ = ['compile']

def fullname(prefix, name):
    return prefix + '_' + name if prefix else name

def fullqualifiedname(prefix, v):
     return fullname(prefix, str(v))

def fullinputname(prefix, v):
    if   v == VCC:
        return '$true'
    elif v == GND:
        return '$false'
    else:
        return fullqualifiedname(prefix, v)

def compileinput(prefix, from_, to):
    assert isinstance(from_, BitType)
    to = fullqualifiedname(prefix, to)
    from_ = fullinputname(prefix, from_)
    return '.names %s %s\n1 1\n' % (from_, to)

def compileinputs(instance, prefix):
    s = ''
    for k, v in instance.interface.ports.items():
        if v.isinput():
            if isinstance(v, ArrayType):
                for i in range(len(v)):
                    w = v[i].value()
                    if w:
                        s += compileinput(prefix, w, v[i])
            else:
                w = v.value()
                if w:
                    s += compileinput(prefix, w, v)
    return s


def compileprimitive(prim, prefix, location):

    assert not isdefinition(prim)

    def arg(k,v): # helper function to format arg
        if not isinstance(v, str): v = str(v)
        return '%s=%s' % (k, v)

    # collect args
    args = []
    for k, v in prim.interface.ports.items():
        if v.direction == INPUT or v.direction == INOUT:
            # connect input to an output by tracing the net
            w = v.value()
            if w:
                w = fullinputname(prefix, w)
                args.append( arg(k,w) )
        elif v.direction == OUTPUT or v.direction == INOUT:
            # connect output to a wire with a fully qualified name
            w = fullqualifiedname(prefix, v)
            args.append( arg(k,w) )

    s  = '.gate ' + str(prim.__class__.__name__) + ' ' + ' '.join(args) + '\n'

    # parameters
    for k, v in prim.kwargs.items():
        if k == 'LUT_INIT':
            if isinstance(v, tuple):
                v = v[0]
            #v = int('0x'+v[4:], 0) # convert from verilog hex
            v = bin(v)[2:]         # convert to binary, skipping 0b
            s += '.param %s %s\n' % (k, v)

    # attributes - only location for now
    if location:
        s += '.attr loc "%d,%d/%d"\n' % location

    return s


def compiledefinition(defn, prefix, location=None):
    s = ''
    for instance in defn.instances:
        wiredefaultclock(defn, instance)
        if location and instance.loc:
            x = location[0] + instance.loc[0]
            y = location[1] + instance.loc[1]
            z = location[2] + instance.loc[2]
            instlocation = (x,y,z)
        else:
            instlocation = None
        s += compileinstance(instance, prefix, instlocation)
    s += '# wire instance outputs\n'
    return s + compileinputs(defn, prefix)

def compileinstance(instance, prefix, location):
    defn = type(instance)
    if isdefinition(defn):
        instname = fullname(prefix, instance.name)
        s  = '# ' + str(defn) + ' ' + instname + ' instances\n'
        s += compiledefinition(defn, instname, location)
        s += '# wire instance inputs\n'
        s += compileinputs(instance, prefix)
        s += '# ' + str(defn) + ' ' + instname + '\n'
    else:
        s  = compileprimitive(instance, prefix, location)
    return s

def compile(main, origin=None):

    s = '.model %s\n' % main.__name__

    inputs = []
    outputs = []
    for k, v in main.interface.ports.items():
        if v.direction == INPUT or v.direction == INOUT:
            inputs.append(k)
        if v.direction == OUTPUT or v.direction == INOUT:
            outputs.append(k)
    s += '.inputs '  + " ".join(outputs) + '\n' # flip inputs
    s += '.outputs ' + " ".join(inputs) + '\n'  # flip outputs
         
    s += '.names $false\n'
    s += '.names $true\n'
    s += '1\n'

    if origin and len(origin) == 2:
        origin = (origin[0], origin[1], 0)
    s += compiledefinition(main, '', origin)

    return s




try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache

cachedFunctions = []


def clear_cachedFunctions():
    for func in cachedFunctions:
        func.cache_clear()


def cache_definition(fn):
    cachedFunctions.append(lru_cache(maxsize=None)(fn))
    return cachedFunctions[-1]

def singleton(cls):
    instance = [None]
    def wrapper(*args, **kwargs):
        if instance[0] is None:
            instance[0] = cls(*args, **kwargs)
        return instance[0]

    return wrapper

# wires
from .port import *
from .wire import *

# types
from .t import *
from .bit import *
from .array import *
from .bits import *
# Define default operators that raise exceptions, rename so it doesn't muck
# with `from magma import *` code that also uses operator
import magma.operators
from .tuple import *
from .clock import *
from .conversions import *
from .interface import *

# circuit
from .circuit import *

# higher-order operators
from .braid import *

# verilog
from .fromverilog import *
from .backend.verilog import *

# compile
from .compile import *

from .logging import warning


mantle_target = None
def set_mantle_target(t):
     global mantle_target
     if mantle_target is not None and mantle_target != t:
         warning('changing mantle target', mantle_target, t )
     mantle_target = t

from .backend.util import set_codegen_debug_info
from .enum import Enum
import magma.util

from .is_primitive import isprimitive
from .is_definition import isdefinition

from .uniquification import UniquificationPass

from .product import Product


from hwtypes.bit_vector_abc import TypeFamily


BitVector = Bits
UIntVector = UInt
SIntVector = SInt
_Family_ = TypeFamily(Bit, BitVector, UIntVector, SIntVector)
def get_family():
    return _Family_



"""
Example usage:
$ MAGMA_LOG_LEVEL=DEBUG MAGMA_LOG_STREAM=stdout pytest
"""
from __future__ import absolute_import
from __future__ import print_function

import logging
import traceback
import inspect
import sys
import os
from io import StringIO
import colorlog


streams = {
    "stdout": sys.stdout,
    "stderr": sys.stderr
}

stream = os.getenv("MAGMA_LOG_STREAM", "stderr")
if stream not in streams:
    logging.warning(f"Unsupported value for MAGMA_LOG_STREAM: {stream} "
                    "using stderr instead")
log_stream = streams.get(stream, sys.stderr)
log = logging.getLogger("magma")
handler = colorlog.StreamHandler(log_stream)
handler.setFormatter(colorlog.ColoredFormatter(
    '%(name)s:%(log_color)s%(levelname)s%(reset)s:%(message)s'))
log.addHandler(handler)


level = os.getenv("MAGMA_LOG_LEVEL", "INFO")
if level in ["DEBUG", "WARN", "INFO"]:
    log.setLevel(getattr(logging, level))
elif level is not None:
    logging.warning(f"Unsupported value for MAGMA_LOG_LEVEL: {level}")


__magma_include_wire_traceback = os.getenv("MAGMA_INCLUDE_WIRE_TRACEBACK", False)


traceback_limit = int(os.getenv("MAGMA_ERROR_TRACEBACK_LIMIT", "5"))


def get_original_wire_call_stack_frame():
    for frame in inspect.stack():
        if sys.version_info < (3, 5):
            function = inspect.getframeinfo(frame[0]).function
        else:
            function = frame.function
        if function not in ["wire", "connect",
                            "get_original_wire_call_stack_frame", "error",
                            "warn", "print_wire_traceback_wrapped",
                            "report_wiring_error", "report_wiring_warning",
                            "__call__", "wireoutputs"]:
            break
    if sys.version_info < (3, 5):
        return frame[0]
    else:
        return frame.frame


def print_wire_traceback(fn):
    def print_wire_traceback_wrapped(*args, **kwargs):
        include_wire_traceback = kwargs.get("include_wire_traceback", False)
        if include_wire_traceback:
            del kwargs["include_wire_traceback"]
        if include_wire_traceback and __magma_include_wire_traceback:
            fn("="*20 + " BEGIN: MAGMA WIRING ERROR TRACEBACK " + "="*20)
            stack_frame = get_original_wire_call_stack_frame()
            with StringIO() as io:
                traceback.print_stack(f=stack_frame, limit=traceback_limit, file=io)
                for line in io.getvalue().splitlines():
                    fn(line)
            fn("="*20 + " END: MAGMA WIRING ERROR TRACEBACK " + "="*20)
        res = fn(*args, **kwargs)
        return res
    return print_wire_traceback_wrapped


@print_wire_traceback
def debug(message, *args, **kwargs):
    log.debug(message, *args, **kwargs)


@print_wire_traceback
def info(message, *args, **kwargs):
    log.info(message, *args, **kwargs)


@print_wire_traceback
def warning(message, *args, **kwargs):
    log.warning(message, *args, **kwargs)


@print_wire_traceback
def error(message, *args, **kwargs):
    log.error(message, *args, **kwargs)


def get_source_line(filename, lineno):
    with open(filename, "r") as f:
        return f.readlines()[lineno - 1]



from __future__ import absolute_import
from __future__ import print_function
from collections import namedtuple, OrderedDict

from mako.template import Template
from pyverilog.vparser.parser import VerilogParser, Node, Input, Output, ModuleDef, Ioport, Port, Decl
import pyverilog.vparser.parser as parser
from pyverilog.dataflow.visit import NodeVisitor
import pyverilog.vparser.ast as pyverilog_ast

from .t import In, Out, InOut
from .bit import Bit, _BitKind
from .bits import Bits, BitsKind
from .array import Array
from .circuit import DeclareCircuit, DefineCircuit, EndDefine

from .passes.tsort import tsort

import logging

logger = logging.getLogger('magma').getChild('from_verilog')


__all__  = ['DeclareFromVerilog']
__all__ += ['DeclareFromVerilogFile']
__all__ += ['DeclareFromTemplatedVerilog']
__all__ += ['DeclareFromTemplatedVerilogFile']
__all__ += ['DefineFromVerilog']
__all__ += ['DefineFromVerilogFile']
__all__ += ['DefineFromTemplatedVerilog']
__all__ += ['DefineFromTemplatedVerilogFile']


class ModuleVisitor(NodeVisitor):
    def __init__(self, shallow):
        self.__shallow = shallow
        self.defns = OrderedDict()
        self.__defn_stack = []
        self.__instances = {}

    def visit_ModuleDef(self, defn):
        if defn.name in self.defns:
            raise Exception(f"Defn with name {defn.name} appears twice")
        self.defns[defn.name] = defn
        if self.__shallow:
            return defn
        # Collect instances in this definition.
        self.__instances[defn] = set()
        self.__defn_stack.append(defn)
        self.generic_visit(defn)
        self.__defn_stack.pop()
        return defn

    def visit_Instance(self, instance):
        if self.__shallow:
            return instance
        defn = self.__defn_stack[-1]
        assert instance not in self.__instances[defn]
        self.__instances[defn].add(instance)
        return instance

    def get_instances(self, defn):
        return self.__instances[defn]

    def sort(self):
        graph = []
        for defn in self.defns.values():
            insts = [inst.module for inst in self.get_instances(defn)]
            graph.append((defn.name, insts))
        sorted_ = tsort(graph)
        defns = OrderedDict()
        for defn_name, _ in sorted_:
            defns[defn_name] = self.defns[defn_name]
        self.defns = defns


def convert(input_type, target_type):
    if isinstance(input_type, _BitKind) and \
       isinstance(target_type, _BitKind) and \
       input_type.direction == target_type.direction:
        return target_type
    if isinstance(input_type, BitsKind) and \
       isinstance(target_type, BitsKind) and \
       input_type.N == target_type.N and \
       input_type.T.direction == target_type.T.direction:
        return target_type
    raise NotImplementedError(f"Conversion between {input_type} and "
                              f"{target_type} not supported")

def get_value(v, param_map):
    if isinstance(v, pyverilog_ast.IntConst):
        return int(v.value)
    if isinstance(v, pyverilog_ast.Rvalue):
        return get_value(v.var, param_map)
    if isinstance(v, (pyverilog_ast.Minus, pyverilog_ast.Uminus)):
        return get_value(v.left, param_map) - get_value(v.right, param_map)
    if isinstance(v, pyverilog_ast.Plus):
        return get_value(v.left, param_map) + get_value(v.right, param_map)
    if isinstance(v, pyverilog_ast.Identifier):
        return param_map[v.name]
    else:
        raise NotImplementedError(type(v))


def get_width(width, param_map):
    msb = get_value(width.msb, param_map)
    lsb = get_value(width.lsb, param_map)
    return msb - lsb + 1


def get_type(io, type_map, param_map):
    if isinstance(io, Input):
        direction = In
    elif isinstance(io, Output):
        direction = Out
    else:
        direction = InOut

    if io.width is None:
        typ = Bit
    else:
        width = get_width(io.width, param_map)
        typ = Bits[width]

    # Generate multidimensional arrays if necessary. Note that we guard with
    # hasattr to make this backwards compatible.
    if hasattr(io, "dimensions") and io.dimensions is not None:
        for length in reversed(io.dimensions.lengths):
            width = get_width(length, param_map)
            typ = Array[width, typ]

    typ = direction(typ)

    if io.name in type_map:
        typ = convert(typ, type_map[io.name])

    return typ


def ParseVerilogModule(node, type_map):
    args = []
    ports = []

    param_map = {}
    for param in node.paramlist.params:
        for p in param.list:
            param_map[p.name] = get_value(p.value, param_map)

    for port in node.portlist.ports:
        if isinstance(port, Ioport):
            io = port.first
            args.append(io.name)
            args.append(get_type(io, type_map, param_map))
        elif isinstance(port, Port):
            ports.append(port.name)
        else:
            raise NotImplementedError(type(port))

    if ports:
        assert not args, "Can we have mixed declared and undeclared types in a Verilog module?"
        for port in ports:
            found = False
            for child in node.children():
                if isinstance(child, Decl):
                    for sub_child in child.children():
                        if isinstance(sub_child, (parser.Input, parser.Output, parser.Inout)) and \
                                sub_child.name == port:
                            args.append(sub_child.name)
                            args.append(get_type(sub_child, type_map, param_map))
                            found = True
                            break
                if found:
                    break
            else:
                raise Exception(f"Could not find type declaration for port {port}")

    return node.name, args


def FromVerilog(source, func, type_map, target_modules=None, shallow=False,
                external_modules={}, param_map={}):
    parser = VerilogParser()
    ast = parser.parse(source)
    visitor = ModuleVisitor(shallow)
    visitor.visit(ast)
    if not shallow:
        visitor.sort()

    def _get_lines(start_line, end_line):
        if shallow:
            return source
        lines = source.split("\n")
        return "\n".join(lines[start_line - 1:end_line])

    if not external_modules.keys().isdisjoint(visitor.defns.keys()):
        intersection = external_modules.keys() & visitor.defns.keys()
        raise Exception(f"Modules defined in both external_modules and in "
                        f"parsed verilog: {intersection}")
    magma_defns = external_modules.copy()
    for name, verilog_defn in visitor.defns.items():
        parsed_name, args = ParseVerilogModule(verilog_defn, type_map)
        assert parsed_name == name
        magma_defn = func(name, *args)
        magma_defn.coreir_config_param_types = param_map
        if func == DefineCircuit:
            # Attach relevant lines of verilog source.
            magma_defn.verilogFile = _get_lines(
                verilog_defn.lineno, verilog_defn.end_lineno)
            if not shallow:
                for instance in visitor.get_instances(verilog_defn):
                    instance_defn = magma_defns[instance.module]
                    instance_defn()
            EndDefine()
        magma_defn.verilog_source = source
        magma_defns[name] = magma_defn

    if len(magma_defns) == 0:
        logger.warning(f"Did not import any modules from verilog, either could "
                       f"not parse or could not find any of the target_modules "
                       f"({target_modules})")
    # Filter back out external modules.
    magma_defns = {name : magma_defns[name] for name in visitor.defns}
    if target_modules is None:
        return list(magma_defns.values())
    # Filter modules based on target_modules list.
    return [v for k, v in magma_defns.items() if k in target_modules]

def FromVerilogFile(file, func, type_map, target_modules=None, shallow=False,
                    external_modules={}, param_map={}):
    if file is None:
        return None
    verilog = open(file).read()
    result = FromVerilog(verilog, func, type_map, target_modules, shallow,
                         external_modules, param_map=param_map)
    # Store the original verilog file name, currently used by m.compile to
    # generate a .sv when compiling a circuit that was defined from a verilog
    # file
    for item in result:
        item.verilog_file_name = file
    return result

def FromTemplatedVerilog(templatedverilog, func, type_map, **kwargs):
    verilog = Template(templatedverilog).render(**kwargs)
    return FromVerilog(verilog, func, type_map)

def FromTemplatedVerilogFile(file, func, type_map, **kwargs):
    if file is None:
        return None
    templatedverilog = open(file).read()
    return FromTemplatedVerilog(templatedverilog, func, type_map, **kwargs)


def DeclareFromVerilog(source, type_map={}, param_map={}):
    return FromVerilog(source, DeclareCircuit, type_map, param_map=param_map)

def DeclareFromVerilogFile(file, target_modules=None, type_map={}, param_map={}):
    return FromVerilogFile(file, DeclareCircuit, type_map, target_modules, param_map=param_map)

def DeclareFromTemplatedVerilog(source, type_map={}, **kwargs):
    return FromTemplatedVerilog(source, DeclareCircuit, type_map, **kwargs)

def DeclareFromTemplatedVerilogFile(file, type_map={}, **kwargs):
    return FromTemplatedVerilogFile(file, DeclareCircuit, type_map, **kwargs)


def DefineFromVerilog(source, type_map={}, target_modules=None, shallow=False,
                      external_modules={}, param_map={}):
    return FromVerilog(source, DefineCircuit, type_map, target_modules,
                       shallow=shallow, external_modules=external_modules,
                       param_map=param_map)

def DefineFromVerilogFile(file, target_modules=None, type_map={},
                          shallow=False, external_modules={}, param_map={}):
    return FromVerilogFile(file, DefineCircuit, type_map, target_modules,
                           shallow=shallow, external_modules=external_modules,
                           param_map=param_map)

def DefineFromTemplatedVerilog(source, type_map={}, **kwargs):
    return FromTemplatedVerilog(source, DefineCircuit, type_map, **kwargs)

def DefineFromTemplatedVerilogFile(file, type_map={}, **kwargs):
    return FromTemplatedVerilogFile(file, DefineCircuit, type_map, **kwargs)



from collections import OrderedDict
from hwtypes import BitVector
import os
from ..bit import VCC, GND, BitType, BitIn, BitOut, MakeBit, BitKind
from ..array import ArrayKind, ArrayType, Array
from ..tuple import TupleKind, TupleType, Tuple
from ..clock import wiredefaultclock, wireclock, ClockType, Clock, ResetType, \
    ClockKind, EnableKind, ResetKind, AsyncResetType, AsyncResetKind, ResetNKind, \
    AsyncResetNKind, AsyncResetNType, ResetType
from ..bitutils import seq2int
from ..backend.verilog import find
from ..logging import error
import coreir
from ..ref import ArrayRef, DefnRef, TupleRef, InstRef
from ..passes import InstanceGraphPass
from ..t import In
import logging
from .util import make_relative, get_codegen_debug_info
from ..interface import InterfaceKind
import inspect
import copy
import json
from .. import singleton
from warnings import warn

from collections import defaultdict

logger = logging.getLogger('magma').getChild('coreir_backend')
level = os.getenv("MAGMA_COREIR_BACKEND_LOG_LEVEL", "WARN")
# TODO: Factor this with magma.logging code for debug level validation
if level in ["DEBUG", "WARN", "INFO"]:
    logger.setLevel(getattr(logging, level))
elif level is not None:
    logger.warning("Unsupported value for MAGMA_COREIR_BACKEND_LOG_LEVEL:"
                   f" {level}")
# logger.setLevel(logging.DEBUG)

class CoreIRBackendError(RuntimeError):
    pass

class keydefaultdict(defaultdict):
    # From https://stackoverflow.com/questions/2912231/is-there-a-clever-way-to-pass-the-key-to-defaultdicts-default-factory
    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError( key )  # pragma: no cover
        else:
            ret = self[key] = self.default_factory(key)
            return ret


def name_to_coreir_select(name):
    if isinstance(name, InstRef):
        return name.inst.name + "." + str(name.name)
    elif isinstance(name, DefnRef):
        return "self." + name.name
    elif isinstance(name, ArrayRef):
        return name_to_coreir_select(name.array.name) + "." + str(name.index)
    elif isinstance(name, TupleRef):
        index = name.index
        try:
            int(index)
            index = f"_{index}"
        except ValueError:
            pass
        return name_to_coreir_select(name.tuple.name) + "." + index
    else:
        raise NotImplementedError(name)


def magma_port_to_coreir(port):
    return name_to_coreir_select(port.name)

# Singleton context meant to be used with coreir/magma code
@singleton
class CoreIRContextSingleton:
    __instance = None

    def get_instance(self):
        return self.__instance

    def reset_instance(self):
        self.__instance = coreir.Context()

    def __init__(self):
        self.__instance = coreir.Context()
CoreIRContextSingleton()

class CoreIRBackend:
    context_to_modules_map = {}
    def __init__(self, context=None, check_context_is_default=True):
        if context is None:
            context = CoreIRContextSingleton().get_instance()
        elif check_context_is_default & (context != CoreIRContextSingleton().get_instance()):
            warn("Creating CoreIRBackend with non-singleton CoreIR context. "
                 "If you're sure you want to do this, set check_context_is_default "
                 "when initializing the CoreIRBackend.")
        if context not in CoreIRBackend.context_to_modules_map:
            CoreIRBackend.context_to_modules_map[context] = {}
        self.modules = CoreIRBackend.context_to_modules_map[context]
        self.context = context
        self.libs = keydefaultdict(self.context.get_lib)
        self.libs_used = set()
        self.__constant_cache = {}
        self.__unique_concat_id = -1

    def check_interface(self, definition):
        # for now only allow Bit, Array, or Tuple
        def check_type(port, errorMessage=""):
            if isinstance(port, ArrayKind):
                check_type(port.T, errorMessage.format("Array({}, {})").format(
                    str(port.N), "{}"))
            elif isinstance(port, TupleKind):
                for (k, t) in zip(port.Ks, port.Ts):
                    check_type(t, errorMessage.format("Tuple({}:{})".format(k, "{}")))
            elif isinstance(port, (BitKind, ClockKind, EnableKind, ResetKind, AsyncResetKind, ResetNKind, AsyncResetNKind)):
                return
            else:
                raise CoreIRBackendError(errorMessage.format(str(port)))
        for name, port in definition.interface.ports.items():
            check_type(type(port), 'Error: Argument {} must be comprised only of Bit, Array, or Tuple')

    def get_type(self, port):
        if isinstance(port, (ArrayType, ArrayKind)):
            _type = self.context.Array(port.N, self.get_type(port.T))
        elif isinstance(port, (TupleType, TupleKind)):
            def to_string(k):
                """
                Unnamed tuples have integer keys (e.g. 0, 1, 2),
                we prefix them with "_" so they can be consumed by coreir's
                Record type (key names are constrained such that they can't be
                integers)
                """
                if isinstance(k, int):
                    return f"_{k}"
                return k
            _type = self.context.Record({
                to_string(k): self.get_type(t) for (k, t) in
                zip(port.Ks, port.Ts)
            })
        elif port.isinput():
            if isinstance(port, (ClockType, ClockKind)):
                _type = self.context.named_types[("coreir", "clk")]
            elif isinstance(port, (AsyncResetType, AsyncResetKind,
                                   AsyncResetNType, AsyncResetNKind)):
                _type = self.context.named_types[("coreir", "arst")]
            else:
                _type = self.context.Bit()
        elif port.isoutput():
            if isinstance(port, (ClockType, ClockKind)):
                _type = self.context.named_types[("coreir", "clkIn")]
            elif isinstance(port, (AsyncResetType, AsyncResetKind,
                                   AsyncResetNType, AsyncResetNKind)):
                _type = self.context.named_types[("coreir", "arstIn")]
            else:
                _type = self.context.BitIn()
        else:
            _type = self.context.BitInOut()
        return _type

    coreirNamedTypeToPortDict = {
        "clk": Clock,
        "coreir.clkIn": Clock
    }

    def get_ports(self, coreir_type, renamed_ports):
        if (coreir_type.kind == "Bit"):
            return BitOut
        elif (coreir_type.kind == "BitIn"):
            return BitIn
        elif (coreir_type.kind == "Array"):
            return Array[len(coreir_type), self.get_ports(coreir_type.element_type, renamed_ports)]
        elif (coreir_type.kind == "Record"):
            elements = {}
            for item in coreir_type.items():
                name = item[0]
                # replace  the in port with I as can't reference that
                if name == "in":
                    name = "I"
                    renamed_ports[name] = "in"
                elements[name] = self.get_ports(item[1], renamed_ports)
            return Tuple(**elements)
        elif (coreir_type.kind == "Named"):
            # exception to handle clock types, since other named types not handled
            if coreir_type.name in self.coreirNamedTypeToPortDict:
                return In(self.coreirNamedTypeToPortDict[coreir_type.name])
            else:
                raise NotImplementedError("not all named types supported yet")
        else:
            raise NotImplementedError("Trying to convert unknown coreir type to magma type")

    def get_ports_as_list(self, ports):
        return [item for i in range(ports.N) for item in [ports.Ks[i], ports.Ts[i]]]

    def convert_interface_to_module_type(self, interface):
        args = OrderedDict()
        for name, port in interface.ports.items():
            args[name] = self.get_type(port)
        return self.context.Record(args)

    def compile_instance(self, instance, module_definition):
        name = instance.__class__.coreir_name
        lib = self.libs[instance.coreir_lib]
        logger.debug(f"Compiling instance {(instance.name, type(instance))}")
        if instance.coreir_genargs is None:
            if hasattr(instance, "wrappedModule") and \
               instance.wrappedModule.context == self.context:
                module = instance.wrappedModule
            else:
                module = lib.modules[name]
            args = {}
            for name, value in instance.kwargs.items():
                if name in {"name", "loc"}:
                    continue  # Skip
                elif isinstance(value, tuple):
                    args[name] = BitVector[value[1]](value[0])
                else:
                    args[name] = value
            args = self.context.new_values(args)
            return module_definition.add_module_instance(instance.name, module, args)
        else:
            generator = lib.generators[name]
            config_args = {}
            for name, value in instance.coreir_configargs.items():
                config_args[name] = value
            config_args = self.context.new_values(config_args)
            gen_args = {}
            for name, value in type(instance).coreir_genargs.items():
                if isinstance(value, AsyncResetKind):
                    value = self.context.named_types["coreir", "arst"]
                elif isinstance(value, ClockKind):
                    value = self.context.named_types["coreir", "clk"]
                gen_args[name] = value
            gen_args = self.context.new_values(gen_args)
            return module_definition.add_generator_instance(instance.name,
                    generator, gen_args, config_args)

    def add_non_input_ports(self, non_input_ports, port):
        if not port.isinput():
            non_input_ports[port] = magma_port_to_coreir(port)
        if isinstance(port, (TupleType, ArrayType)):
            for element in port:
                self.add_non_input_ports(non_input_ports, element)

    def compile_declaration(self, declaration):
        if declaration.coreir_lib is not None:
            self.libs_used.add(declaration.coreir_lib)
        # These libraries are already available by default in coreir, so we
        # don't need declarations
        if declaration.coreir_lib in ["coreir", "corebit", "commonlib"]:
            if declaration.coreir_genargs is None:
                return self.libs[declaration.coreir_lib].modules[declaration.coreir_name]
            else:
                return self.libs[declaration.coreir_lib].generators[declaration.coreir_name]
        if declaration.name in self.modules:
            logger.debug(f"    {declaration} already compiled, skipping")
            return
        module_type = self.convert_interface_to_module_type(declaration.interface)
        if isinstance(declaration.interface, InterfaceKind):
            module_type = self.context.Flip(module_type)

        kwargs = {}
        if hasattr(declaration, "coreir_config_param_types"):
            kwargs["cparams"] = \
                self.make_cparams(declaration.coreir_config_param_types)

        coreir_module = \
            self.context.global_namespace.new_module(declaration.coreir_name,
                                                     module_type, **kwargs)
        if get_codegen_debug_info() and declaration.debug_info:
            coreir_module.add_metadata("filename", json.dumps(make_relative(declaration.debug_info.filename)))
            coreir_module.add_metadata("lineno", json.dumps(str(declaration.debug_info.lineno)))
        return coreir_module

    def compile_definition_to_module_definition(self, definition, module_definition):
        if definition.coreir_lib is not None:
            self.libs_used.add(definition.coreir_lib)
        non_input_ports = {}
        for name, port in definition.interface.ports.items():
            logger.debug("{}, {}, {}".format(name, port, port.isoutput()))
            self.add_non_input_ports(non_input_ports, port)

        for instance in definition.instances:
            wiredefaultclock(definition, instance)
            wireclock(definition, instance)
            coreir_instance = self.compile_instance(instance, module_definition)
            if get_codegen_debug_info() and getattr(instance, "debug_info", False):
                coreir_instance.add_metadata("filename", json.dumps(make_relative(instance.debug_info.filename)))
                coreir_instance.add_metadata("lineno", json.dumps(str(instance.debug_info.lineno)))
            for name, port in instance.interface.ports.items():
                self.add_non_input_ports(non_input_ports, port)

        for instance in definition.instances:
            for name, port in instance.interface.ports.items():
                self.connect_non_outputs(module_definition, port,
                                         non_input_ports)

        for port in definition.interface.ports.values():
            self.connect_non_outputs(module_definition, port, non_input_ports)

    def connect_non_outputs(self, module_definition, port,
                      non_input_ports):
        # Recurse into non input types that may contain inout children
        if isinstance(port, TupleType) and not port.isinput() or \
                isinstance(port, ArrayType) and not port.T.isinput():
            for elem in port:
                self.connect_non_outputs(module_definition, elem,
                                         non_input_ports)
        elif not port.isoutput():
            self.connect(module_definition, port, port.value(),
                         non_input_ports)

    def make_cparams(self, param_types):
        cparams = {}
        for param, type_ in param_types.items():
            if type_ is int:
                type_ = self.context.Int()
            elif type_ is bool:
                type_ = self.context.Bool()
            elif type_ is str:
                type_ = self.context.String()
            elif type_ is BitVector:
                type_ = self.context.BitVector()
            else:
                raise NotImplementedError(type_)
            cparams[param] = type_
        return self.context.newParams(cparams)

    def compile_definition(self, definition):
        logger.debug(f"Compiling definition {definition}")
        if definition.name in self.modules:
            logger.debug(f"    {definition} already compiled, skipping")
            return self.modules[definition.name]
        self.check_interface(definition)
        module_type = self.convert_interface_to_module_type(definition.interface)
        kwargs = {}
        if hasattr(definition, "coreir_config_param_types"):
            kwargs["cparams"] = \
                self.make_cparams(definition.coreir_config_param_types)

        coreir_module = \
            self.context.global_namespace.new_module(definition.coreir_name,
                                                     module_type, **kwargs)
        if get_codegen_debug_info() and definition.debug_info:
            coreir_module.add_metadata("filename", json.dumps(make_relative(definition.debug_info.filename)))
            coreir_module.add_metadata("lineno", json.dumps(str(definition.debug_info.lineno)))

        # If this module was imported from verilog, do not go through the
        # general module construction flow. Instead just attach the verilog
        # source as metadata and return the module. Also, we attach any
        # contained instances as CoreIR instances.
        if hasattr(definition, "verilogFile") and definition.verilogFile:
            verilog_metadata = {"verilog_string": definition.verilogFile}
            coreir_module.add_metadata("verilog", json.dumps(verilog_metadata))
            module_definition = coreir_module.new_definition()
            coreir_module.definition = module_definition
            for instance in definition.instances:
                self.compile_instance(instance, module_definition)
            return coreir_module

        module_definition = coreir_module.new_definition()
        self.compile_definition_to_module_definition(definition, module_definition)
        coreir_module.definition = module_definition
        return coreir_module

    def connect(self, module_definition, port, value, non_input_ports):
        self.__unique_concat_id

        # allow clocks or arrays of clocks to be unwired as CoreIR can wire them up
        def is_clock_or_nested_clock(p):
            if isinstance(p, (ClockType, ClockKind)):
                return True
            elif isinstance(p, (ArrayType, ArrayKind)):
                return is_clock_or_nested_clock(p.T)
            elif isinstance(p, (TupleType, TupleKind)):
                for item in p.Ts:
                    if is_clock_or_nested_clock(item):
                        return True
            return False

        if value is None and is_clock_or_nested_clock(port):
            return
        elif value is None:
            if port.isinout():
                # Skip inouts because they might be connected as an input
                return
            raise Exception(f"Got None for port '{port.debug_name}', is it "
                            "connected to anything?")
        elif isinstance(value, coreir.Wireable):
            source = value

        elif isinstance(value, ArrayType) and all(x in {VCC, GND} for x in value):
            source = self.get_constant_instance(value, len(value),
                    module_definition)
        elif value.anon() and isinstance(value, ArrayType):
            for p, v in zip(port, value):
                self.connect(module_definition, p, v, non_input_ports)
            return
        elif isinstance(value, TupleType) and value.anon():
            for p, v in zip(port, value):
                self.connect(module_definition, p, v, non_input_ports)
            return
        elif value is VCC or value is GND:
            source = self.get_constant_instance(value, None, module_definition)
        else:
            # logger.debug((value, non_input_ports))
            # logger.debug((id(value), [id(key) for key in non_input_ports]))
            source = module_definition.select(non_input_ports[value])
        sink = module_definition.select(magma_port_to_coreir(port))
        module_definition.connect(source, sink)
        if get_codegen_debug_info() and getattr(port, "debug_info", False):
            module_definition.add_metadata(source, sink, "filename", json.dumps(make_relative(port.debug_info.filename)))
            module_definition.add_metadata(source, sink, "lineno", json.dumps(str(port.debug_info.lineno)))


    __unique_constant_id = -1
    def get_constant_instance(self, constant, num_bits, module_definition):
        if module_definition not in self.__constant_cache:
            self.__constant_cache[module_definition] = {}

        bit_type_to_constant_map = {
            GND: 0,
            VCC: 1
        }
        if constant in bit_type_to_constant_map:
            value = bit_type_to_constant_map[constant]
        elif isinstance(constant, ArrayType):
            value = BitVector[len(constant)]([bit_type_to_constant_map[x] for x in constant])
        else:
            raise NotImplementedError(constant)
        if (value, num_bits) not in self.__constant_cache[module_definition]:
            self.__unique_constant_id += 1
            if num_bits is None:
                config = self.context.new_values({"value": bool(value)})
                name = "bit_const_{}_{}".format(value, num_bits)
                corebit_const_module = self.libs['corebit'].modules["const"]
                module_definition.add_module_instance(name, corebit_const_module, config)
            else:
                gen_args = self.context.new_values({"width": num_bits})
                config = self.context.new_values({"value": value})
                # name = "const_{}_{}".format(constant, self.__unique_constant_id)
                name = "const_{}_{}".format(value, num_bits)
                const_generator = self.libs['coreir'].generators["const"]
                module_definition.add_generator_instance(name, const_generator, gen_args, config)
            # return module_definition.select("{}.out".format(name))
            self.__constant_cache[module_definition][(value, num_bits)] = module_definition.select("{}.out".format(name))
        return self.__constant_cache[module_definition][(value, num_bits)]

    def compile_dependencies(self, defn):
        pass_ = InstanceGraphPass(defn)
        pass_.run()
        dependency_names = [key.name for key, _ in pass_.tsortedgraph]
        logger.debug(f"tsortedgraph: {dependency_names}")
        for key, _ in pass_.tsortedgraph:
            if key == defn:
                continue
            if key.name in self.modules:
                continue
            if key.is_definition:
                # don't try to compile if already have definition
                if hasattr(key, 'wrappedModule') and \
                   key.wrappedModule.context == self.context:
                    self.modules[key.name] = key.wrappedModule
                    self.libs_used |= key.coreir_wrapped_modules_libs_used
                else:
                    self.modules[key.name] = self.compile_definition(key)
                    key.wrappedModule = self.modules[key.name]
                    key.coreir_wrapped_modules_libs_used = \
                        copy.copy(self.libs_used)
                    logger.debug(f"Compiled module {key.name}, libs used = {self.libs_used}")
            else:
                self.modules[key.name] = self.compile_declaration(key)

    def compile(self, defn_or_declaration):
        logger.debug(f"Compiling: {defn_or_declaration.name}")
        if defn_or_declaration.is_definition:
            self.compile_dependencies(defn_or_declaration)
            # don't try to compile if already have definition
            if hasattr(defn_or_declaration, 'wrappedModule') and \
               defn_or_declaration.wrappedModule.context == self.context:
                self.modules[defn_or_declaration.name] = defn_or_declaration.wrappedModule
                self.libs_used |= defn_or_declaration.coreir_wrapped_modules_libs_used
            else:
                self.modules[defn_or_declaration.name] = self.compile_definition(defn_or_declaration)
                defn_or_declaration.wrappedModule = self.modules[defn_or_declaration.name]
                defn_or_declaration.coreir_wrapped_modules_libs_used = \
                    copy.copy(self.libs_used)
        else:
            self.modules[defn_or_declaration.name] = self.compile_declaration(defn_or_declaration)
        return self.modules

    def flatten_and_save(self, module, filename, namespaces=["global"], flatten=True, verifyConnectivity=True):
        passes = ["rungenerators", "wireclocks-coreir"]
        if verifyConnectivity:
            passes += ["verifyconnectivity --noclkrst"]
        if flatten:
            passes += ["flattentypes", "flatten"]
        self.context.run_passes(passes, namespaces)
        module.save_to_file(filename)

def compile(main, file_name=None, context=None, check_context_is_default=True):
    backend = CoreIRBackend(context, check_context_is_default)
    backend.compile(main)
    if file_name is not None:
        return backend.modules[main.coreir_name].save_to_file(file_name)
    else:
        return backend.modules[main.coreir_name]



import ast
import textwrap
import sys
import six
import inspect
from functools import wraps
if sys.version_info > (3, 0):
    from functools import reduce
from . import cache_definition
import operator
from collections import namedtuple, Counter
from .clock import ClockTypes
from .interface import *
from .wire import *
from .t import Flip
from .array import ArrayType
from .tuple import TupleType
from .bit import VCC, GND
from .config import get_debug_mode
from .debug import get_callee_frame_info, debug_info
from .logging import warning
from .port import report_wiring_warning, report_wiring_error
from .is_definition import isdefinition


__all__  = ['AnonymousCircuitType']
__all__ += ['AnonymousCircuit']

__all__ += ['CircuitType']
__all__ += ['Circuit']
__all__ += ['DeclareCircuit']
__all__ += ['DefineCircuit', 'EndDefine', 'EndCircuit']
__all__ += ['getCurrentDefinition']

__all__ += ['CopyInstance']
__all__ += ['circuit_type_method']
__all__ += ['circuit_generator']


circuit_type_method = namedtuple('circuit_type_method', ['name', 'definition'])


def circuit_to_html(cls):
    if isdefinition(cls):
        # Avoid circular dependency so dot backend can use passes
        from .backend.dot import to_html
        return to_html(cls)
    else:
        return repr(cls)

# create an attribute for each port
def setports(self, ports):
    #print('setports', ports)
    for name, port in ports.items():
        #print(self, port, type(port))
        if isinstance(name, str):
            setattr(self, name, port)

#
# Metaclass for creating circuits
#
class CircuitKind(type):

    def __new__(metacls, name, bases, dct):
        #print('CircuitKind new:', name)

        # override circuit class name
        name = dct.setdefault('name', name)

        dct.setdefault('renamed_ports', {})
        dct.setdefault('primitive', False)
        dct.setdefault('coreir_lib', 'global')

        if get_debug_mode():
            if not dct.get("debug_info", False):
                callee_frame = inspect.getframeinfo(inspect.currentframe().f_back.f_back)
                module = inspect.getmodule(inspect.stack()[2][0])
                dct["debug_info"] = debug_info(callee_frame.filename,
                                               callee_frame.lineno, module)
        else:
            dct["debug_info"] = None


        # create a new circuit class
        cls = type.__new__(metacls, name, bases, dct)

        for method in dct.get('circuit_type_methods', []):
            setattr(cls, method.name, method.definition)

        # create interface for this circuit class
        if hasattr(cls, 'IO') and not isinstance(cls.IO, InterfaceKind):
            # turn IO attribite into an Interface
            cls.IO = DeclareInterface(*cls.IO)
            cls.interface = cls.IO(defn=cls, renamed_ports=dct["renamed_ports"])
            setports(cls, cls.interface.ports)

        return cls

    def __call__(cls, *largs, **kwargs):
        #print('CircuitKind call:', largs, kwargs)
        if get_debug_mode():
            debug_info = get_callee_frame_info()
            kwargs["debug_info"] = debug_info
        self = super(CircuitKind, cls).__call__(*largs, **kwargs)

        # instance interface for this instance
        if hasattr(cls, 'IO'):
            self.setinterface(cls.IO(inst=self, renamed_ports=cls.renamed_ports))

        return self

    def __str__(cls):
        interface = ""
        if hasattr(cls, "interface"):
            interface = ", ".join(f"{name}: {_type}" for name, _type in cls.interface.items())
            interface = f"({interface})"
        return f"{cls.__name__}{interface}"

    def __repr__(cls):
        if not hasattr(cls, 'IO'):
            return super().__repr__()

        name = cls.__name__
        args = str(cls.IO)
        if hasattr(cls,"instances"):
            s = '{} = DefineCircuit("{}", {})\n'.format(name, name, args)

            sorted_instances = sorted(cls.instances, key=lambda x : x.name)
            # emit instances
            for instance in sorted_instances:
                s += repr(instance) + '\n'

            # emit wires from instances
            for instance in sorted_instances:
                s += repr(instance.interface)

            # for input in cls.interface.inputs():
            s += repr( cls.interface )

            s += "EndCircuit()"
        else:
            s = '{} = DeclareCircuit("{}", {})'.format(name, name, args)

        return s

    def _repr_html_(cls):
        return circuit_to_html(cls)

    def rename(cls, new_name):
        if cls.verilogFile:
            raise Exception("Can not rename a verilog wrapped file")

        old_name = cls.name
        cls.name = new_name
        cls.coreir_name = new_name
        cls.verilog_name = new_name
        cls.__name__ = new_name

        # NOTE(rsetaluri): This is a very hacky way to try to rename wrapped
        # verilog. We simply replace the first instance of "module <old_name>"
        # with "module <new_name>". This ignores the possibility of "module
        # <new_name>" existing anywhere else, most likely in comments etc. The
        # more robust way to do this would to modify the AST directly and
        # generate the new verilog code.
        #
        # Instead, at the top of this function, we raise an exception for
        # verilog wrapped files. That is, for now it is considered an error to
        # try to rename a verilog wrapped circuit.
        if cls.verilogFile:
            find_str = f"module {old_name}"
            replace_str = f"module {new_name}"
            assert cls.verilogFile.find(find_str) != -1
            cls.verilogFile = cls.verilogFile.replace(find_str, replace_str, 1)

    def find(cls, defn):
        name = cls.__name__
        if not isdefinition(cls):
            return defn
        for i in cls.instances:
            type(i).find(defn)
        if name not in defn:
            defn[name] = cls
        return defn

#
# Abstract base class for circuits
#
@six.add_metaclass(CircuitKind)
class AnonymousCircuitType(object):

    def __init__(self, *largs, **kwargs):

        self.kwargs = dict(**kwargs)
        if "debug_info" in self.kwargs:
            # Not an instance parameter, internal debug argument
            del self.kwargs["debug_info"]
        if hasattr(self, 'default_kwargs'):
            for key in self.default_kwargs:
                if key not in kwargs:
                    self.kwargs[key] = self.default_kwargs[key]

        self.name = kwargs['name'] if 'name' in kwargs else ""

        self.loc = kwargs['loc'] if 'loc' in kwargs else None
        if self.loc and len(self.loc) == 2:
            self.loc = (self.loc[0], self.loc[1], 0)

        self.interface = None
        self.defn = None
        self.used = False
        self.is_instance = True

        self.debug_info = kwargs.get("debug_info", None)

    def set_debug_info(self, debug_info):
        self.debug_info = debug_info

    def __str__(self):
        if self.name:
            return f"{self.name}<{type(self)}>"
        else:
            name = f"AnonymousCircuitInst{id(self)}"
            interface = ""
            interface = ", ".join(f"{name}: {type(value)}" for name, value in self.interface.ports.items())
            return f"{name}<{interface}>"

    def __repr__(self):
        args = []
        for k, v in self.interface.ports.items():
            args.append('"{}"'.format(k))
            args.append(repr(v))
        if self.name:
            return '{} = {}({})'.format(self.name, type(self).__name__, ', '.join(args))
        else:
            return '{}({})'.format(type(self).__name__, ', '.join(args))

        #return '{} = {}({})  # {} {}'.format(str(self), str(type(self)),
        #    ', '.join(args), self.filename, self.lineno)

    def _repr_html_(self):
        return circuit_to_html(self)

    def __getitem__(self, key):
        return self.interface[key]

    # wire a list of outputs to the circuit's inputs
    def wireoutputs(self, outputs, debug_info):
        inputs = self.interface.inputs()
        ni = len(inputs)
        no = len(outputs)
        if ni != no:
            report_wiring_warning(f"Number of inputs is not equal to the number of outputs, expected {ni} inputs, got {no}. Only {min(ni,no)} will be wired.",  # noqa
                    debug_info)
        for i in range(min(ni,no)):
            wire(outputs[i], inputs[i], debug_info)

    # wire a single output to the circuit's inputs
    def wire(self, output, debug_info):

        if hasattr(output, 'interface'):
            # wire the circuit's outputs to this circuit's inputs
            self.wireoutputs(output.interface.outputs(), debug_info)
        else:
            # wire the output to this circuit's input (should only have 1 input)
            inputs = self.interface.inputs()
            ni = len(inputs)
            if ni == 0:
                report_wiring_warning("Wiring an output to a circuit with no input arguments, skipping", debug_info)
                return
            if ni != 1:
                report_wiring_warning(f"Wiring an output to a circuit with more than one input argument, using the first input {inputs[0].debug_name}", debug_info)
            inputs[0].wire( output, debug_info )


    @property
    def debug_name(self):
        defn_str = ""
        if hasattr(self, 'defn') and self.defn is not None:
            defn_str = str(self.defn.name)
        return f"{defn_str}.{self.name}"

    def __call__(input, *outputs, **kw):
        if get_debug_mode():
            debug_info = get_callee_frame_info()
        else:
            debug_info = None

        no = len(outputs)
        if len(outputs) == 1:
            input.wire(outputs[0], debug_info)
        elif len(outputs) >= 1:  # In case there are only kw
            input.wireoutputs(outputs, debug_info)

        # wire up extra arguments, name to name
        #
        # this code should be changed to use clock types ...
        #
        for key, value in kw.items():
            if key == 'enable': key = 'CE'
            if key == 'reset':  key = 'RESET'
            if key == 'set':    key = 'SET' # NYI
            if key == 'ce':     key = 'CE'  # depreciated
            if hasattr(input, key):
                i = getattr(input, key)
                wire( value, getattr(input, key), debug_info)
            else:
                report_wiring_warning('Instance {} does not have input {}'.format(input.debug_name, key), debug_info)

        o = input.interface.outputs()
        return o[0] if len(o) == 1 else tuple(o)

    def setinterface(self, interface):
        setports(self, interface.ports)
        self.interface = interface
        return self

    def on(self):
        self.used = True
        return self

    def off(self):
        self.used = False
        return self

    def rename(self, name):
        self.name = name
        return self

    def isclocked(self):
        return self.interface.isclocked()

    def clockargs(self):
        return self.interface.clockargs()

    def inputargs(self):
        return self.interface.inputargs()

    def outputargs(self):
        return self.interface.outputargs()

#
# AnonymousCircuits are like macros - the circuit instances are not placed
#
def AnonymousCircuit(*decl):
    if len(decl) == 1:
        decl = decl[0]
    return AnonymousCircuitType().setinterface(Interface(decl))


#
# Placed circuit - instances placed in a definition
#
class CircuitType(AnonymousCircuitType):
    def __init__(self, *largs, **kwargs):
        super(CircuitType, self).__init__(*largs, **kwargs)

        # Circuit instances are placed if within a definition
        global currentDefinition
        if currentDefinition:
             currentDefinition.place(self)

    def __repr__(self):
        args = []
        for k, v in self.kwargs.items():
            if isinstance(v, tuple):
                 # {   # Format identifier
                 # 0:  # first parameter
                 # #   # use "0x" prefix
                 # 0   # fill with zeroes
                 # {1} # to a length of n characters (including 0x), defined by the second parameter
                 # x   # hexadecimal number, using lowercase letters for a-f
                 # }   # End of format identifier
                 if len(v) == 2:
                     v = "{0:#0{1}x}".format(v[0], v[1] // 4)
            else:
                 v = '"{}"'.format(v)
            args.append("%s=%s"%(k, v))
        if self.name:
            return '{} = {}({})'.format(self.name, type(self).__name__, ', '.join(args))
        else:
            return '{}({})'.format(type(self).__name__, ', '.join(args))

        #return '{} = {}({})  # {} {}'.format(str(self), str(type(self)),
        # cls.filename, cls.lineno)

# DeclareCircuit Factory
def DeclareCircuit(name, *decl, **args):
    if get_debug_mode():
        debug_info = get_callee_frame_info()
    else:
        debug_info = None
    dct = dict(
        IO=decl,
        debug_info=debug_info,
        is_definition=False,
        primitive=args.get('primitive', True),
        stateful=args.get('stateful', False),
        simulate=args.get('simulate'),
        firrtl_op=args.get('firrtl_op'),
        circuit_type_methods=args.get('circuit_type_methods', []),
        coreir_lib=args.get('coreir_lib', "global"),
        coreir_name=args.get('coreir_name', name),
        coreir_genargs=args.get('coreir_genargs', None),
        coreir_configargs=args.get('coreir_configargs', {}),
        verilog_name=args.get('verilog_name', name),
        default_kwargs=args.get('default_kwargs', {}),
        renamed_ports=args.get('renamed_ports', {})
    )
    return CircuitKind( name, (CircuitType,), dct )



# Maintain a current definition and stack of nested definitions

currentDefinition = None
currentDefinitionStack = []

def getCurrentDefinition():
    global currentDefinition
    return currentDefinition

def pushDefinition(defn):
    global currentDefinition
    if currentDefinition:
        currentDefinitionStack.append(currentDefinition)
    currentDefinition = defn

def popDefinition():
    global currentDefinition
    if len(currentDefinitionStack) > 0:
        currentDefinition = currentDefinitionStack.pop()
    else:
        currentDefinition = None

class DefineCircuitKind(CircuitKind):
    def __new__(metacls, name, bases, dct):

        if 'name' not in dct:
            # Check if we are a subclass of something other than Circuit
            for base in bases:
                if base is not Circuit:
                    if not issubclass(base, Circuit):
                        raise Exception("Must subclass from Circuit or a "
                                "subclass of Circuit. {}".format(base))
                    # If so, we will inherit the name of the first parent
                    dct['name'] = base.name
                    break
            else:
                dct['name'] = name
        name = dct['name']
        dct["renamed_ports"] = dct.get("renamed_ports", {})

        self = CircuitKind.__new__(metacls, name, bases, dct)

        self.verilog = None
        self.verilogFile = None
        self.verilogLib = None

        self.verilog_name = dct.get('verilog_name', name)
        self.coreir_name = dct.get('coreir_name', name)
        self.coreir_genargs = dct.get('coreir_genargs', None)
        self.coreir_configargs = dct.get('coreir_configargs', {})
        self.default_kwargs = dct.get('default_kwargs', {})

        self.firrtl = None

        self._instances = []
        self.instanced_circuits_counter = Counter()
        self.instance_name_counter = Counter()
        self.instance_name_map = {}
        self._is_definition = dct.get('is_definition', False)
        self.is_instance = False

        if hasattr(self, 'IO'):

            # create circuit definition
            if hasattr(self, 'definition'):
                 pushDefinition(self)
                 self.definition()
                 self.check_unconnected()
                 self._is_definition = True
                 EndCircuit()

        return self

    def check_unconnected(self):
        for port in self.interface.ports.values():
            if issubclass(type(port), ClockTypes):
                continue
            if port.isinput() and not port.driven():
                report_wiring_error(f"Output port {self.name}.{port.name} not driven", self.debug_info)

        for inst in self.instances:
            for port in inst.interface.ports.values():
                if issubclass(type(port), ClockTypes):
                    continue
                if port.isinput() and not port.driven():
                    report_wiring_error(f"Input port {inst.name}.{port.name} not driven", inst.debug_info)

    @property
    def is_definition(self):
        return self._is_definition or self.verilog or self.verilogFile

    @property
    def instances(self):
        return self._instances

    def inspect_name(cls, inst):
        # Try to fetch instance name
        with open(inst.debug_info.filename, "r") as f:
            line = f.read().splitlines()[inst.debug_info.lineno - 1]
            tree = ast.parse(textwrap.dedent(line)).body[0]
            # Simple case when <Name> = <Instance>()
            if isinstance(tree, ast.Assign) and len(tree.targets) == 1 \
                    and isinstance(tree.targets[0], ast.Name):
                name = tree.targets[0].id
                # Handle case when we've seen a name multiple times
                # (e.g. reused inside a loop)
                if cls.instance_name_counter[name] == 0:
                    inst.name = name
                    cls.instance_name_counter[name] += 1
                else:
                    if cls.instance_name_counter[name] == 1:
                        # Append `_0` to the first instance with this
                        # name
                        orig = cls.instance_name_map[name]
                        orig.name += "_0"
                        del cls.instance_name_map[name]
                        cls.instance_name_map[orig.name] = orig
                    inst.name = f"{name}_{cls.instance_name_counter[name]}"
                    cls.instance_name_counter[name] += 1

    #
    # place a circuit instance in this definition
    #
    def place(cls, inst):
        if not inst.name:
            if get_debug_mode():
                cls.inspect_name(inst)
            if not inst.name:
                # Default name if we could not find one or debug mode is off
                inst.name = f"{type(inst).name}_inst{str(cls.instanced_circuits_counter[type(inst).name])}"
                cls.instanced_circuits_counter[type(inst).name] += 1
                cls.instance_name_counter[inst.name] += 1
        else:
            cls.instance_name_counter[inst.name] += 1
        cls.instance_name_map[inst.name] = inst
        inst.defn = cls
        if get_debug_mode():
            inst.stack = inspect.stack()
        cls.instances.append(inst)


# Register graphviz repr if running in IPython.
# There's a bug in IPython which breaks visual reprs
# on types.
try:
    ip = get_ipython()
    html_formatter = ip.display_formatter.formatters['text/html']
    html_formatter.for_type(DefineCircuitKind, circuit_to_html)
    html_formatter.for_type(CircuitKind, circuit_to_html)
except NameError:
    # Not running in IPython right now?
    pass


@six.add_metaclass(DefineCircuitKind)
class Circuit(CircuitType):
    pass


# DefineCircuit Factory
def DefineCircuit(name, *decl, **args):
    if get_debug_mode():
        debug_info = get_callee_frame_info()
    else:
        debug_info = None
    global currentDefinition
    if currentDefinition:
        currentDefinitionStack.append(currentDefinition)

    dct = dict(IO             = decl,
               is_definition  = True,
               primitive      = args.get('primitive', False),
               stateful       = args.get('stateful', False),
               simulate       = args.get('simulate'),
               debug_info     = debug_info,
               verilog_name   = args.get('verilog_name', name),
               coreir_name    = args.get('coreir_name', name),
               coreir_lib     = args.get('coreir_lib', "global"),
               coreir_genargs = args.get('coreir_genargs', None),
               coreir_configargs = args.get('coreir_configargs', None),
               default_kwargs = args.get('default_kwargs', {}),
               renamed_ports = args.get('renamed_ports', {}))

    currentDefinition = DefineCircuitKind( name, (Circuit,), dct)
    return currentDefinition

def EndDefine():
    if currentDefinition:
        debug_info = get_callee_frame_info()
        currentDefinition.end_circuit_filename = debug_info[0]
        currentDefinition.end_circuit_lineno   = debug_info[1]
        popDefinition()
    else:
        raise Exception("EndDefine called without Define/DeclareCircuit")

EndCircuit = EndDefine

def CopyInstance(instance):
    circuit = type(instance)
    new_instance = circuit()
    new_instance.kwargs = instance.kwargs
    return new_instance

def hex(i):
    if i < 10: return chr(ord('0')+i)
    else:      return chr(ord('A')+i-10)


def hstr(init, nbits):
    bits = 1 << int(nbits)
    format = "0x"
    nformat = []
    for i in range(bits//4):
        nformat.append(init%16)
        init //= 16
    nformat.reverse()
    if nformat:
        return format + reduce(operator.add, map(hex, nformat))
    return format


GeneratorArguments = namedtuple('GeneratorArguments', ['args', 'kwargs'])


def circuit_generator(func):
    @cache_definition
    @wraps(func)
    def wrapped(*args, **kwargs):
        result = func(*args, **kwargs)
        # Store arguments to generate the circuit
        result._generator_arguments = GeneratorArguments(args, kwargs)
        return result
    return wrapped

from magma.syntax.combinational import combinational
from magma.syntax.sequential import sequential



import magma as m


def BitOrBits(width):
    if width is None:
        return m.Bit
    if not isinstance(width, int):
        raise ValueError(f"Expected width to be None or int, got {width}")
    return m.Bits[width]


def pretty_str(t):
    if isinstance(t, m.TupleKind):
        args = []
        for i in range(t.N):
            key_str = str(t.Ks[i])
            val_str = pretty_str(t.Ts[i])
            indent = " " * 4
            val_str = f"\n{indent}".join(val_str.splitlines())
            args.append(f"{key_str} = {val_str}")
        # Pretty print by using newlines + indent
        joiner = ",\n    "
        result = joiner.join(args)
        # Insert first newline + indent and last newline
        result = "\n    " + result + "\n"
        s = f"Tuple({result})"
    elif isinstance(t, m.BitsKind):
        s = str(t)
    elif isinstance(t, m.ArrayKind):
        s = f"Array[{t.N}, {pretty_str(t.T)}]"
    else:
        s = str(t)
    return s


def pretty_print_type(t):
    print(pretty_str(t))



import sys
import os
import inspect
import subprocess

from .passes import DefinitionPass, InstanceGraphPass
from .backend import verilog, blif, firrtl, dot
from .config import get_compile_dir
from .logging import warning
from .uniquification import uniquification_pass, UniquificationMode
import magma as m


__all__ = ['compile']


def write_file(file_name, extension, code):
    with open("{}.{}".format(file_name, extension), 'w') as file:
        file.write(code)


def __compile_to_coreir(main, file_name, opts):
    # Underscore so our coreir module doesn't conflict with coreir bindings
    # package.
    from .frontend import coreir_
    backend = coreir_.GetCoreIRBackend()
    backend.compile(main)
    passes = opts.get("passes", [])
    if "markdirty" not in passes:
        passes.append("markdirty")
    namespaces = opts.get("namespaces", ["global"])
    backend.context.run_passes(passes, namespaces)

    backend.modules[main.coreir_name].save_to_file(file_name + ".json")
    if opts.get("output_verilog", False):
        deps = opts.get("coreir_libs", set())
        pass_ = InstanceGraphPass(main)
        pass_.run()
        for key, _ in pass_.tsortedgraph:
            if key.coreir_lib:
                deps.add(key.coreir_lib)
            elif hasattr(key, 'wrappedModule'):
                deps |= key.coreir_wrapped_modules_libs_used
        for namespace in namespaces:
            deps.add(namespace)
        # TODO(rsetaluri): Expose compilation to verilog in pycoreir rather than
        # calling the binary from the command line.
        lib_arg = ""
        if deps:
            lib_arg = f"-l {','.join(deps)}"
        cmd = f"coreir {lib_arg} -i {file_name}.json"
        if opts.get("split", ""):
            split = opts["split"]
            cmd += f" -o \"{split}/*.v\" -s"
        else:
            cmd += f" -o {file_name}.v"
        if opts.get("inline", False):
            cmd += " --inline"
        if opts.get("verilator_debug", False):
            cmd += " --verilator_debug"
        assert not subprocess.run(cmd, shell=True).returncode, \
            "Running coreir failed"


def compile(basename, main, output='coreir-verilog', **kwargs):
    opts = kwargs.copy()

    # Rather than having separate logic for 'coreir-verilog' mode, we defer to
    # 'coreir' mode with the 'output_verilog' option set to True.
    if output == 'coreir-verilog':
        opts["output_verilog"] = True
        output = "coreir"

    # Allow the user to pass in a mode for uniquification. The input is expected
    # as a string and maps to the UniquificationMode enum.
    uniquification_mode_str = opts.get("uniquify", "UNIQUIFY")
    uniquification_mode = getattr(UniquificationMode, uniquification_mode_str,
                                  None)
    if uniquification_mode is None:
        raise ValueError(f"Invalid uniquification mode "
                         f"{uniquification_mode_str}")
    uniquification_pass(main, uniquification_mode)

    if get_compile_dir() == 'callee_file_dir':
        (_, filename, _, _, _, _) = inspect.getouterframes(inspect.currentframe())[1]
        file_path = os.path.dirname(filename)
        file_name = os.path.join(file_path, basename)
    else:
        file_name = basename

    if output == 'verilog':
        suffix = "v"
        # Handle the case when DefineFromVerilogFile is used with a system
        # verilog file
        if hasattr(main, "verilog_file_name") and \
                os.path.splitext(main.verilog_file_name)[-1] == ".sv":
            suffix = "sv"
        write_file(file_name, suffix, verilog.compile(main))
    elif output == 'blif':
        write_file(file_name, 'blif', blif.compile(main))
    elif output == 'firrtl':
        write_file(file_name, 'fir', firrtl.compile(main))
    elif output == 'coreir':
        __compile_to_coreir(main, file_name, opts)
    elif output == 'dot':
        write_file(file_name, 'dot', dot.compile(main))
    else:
        raise NotImplementedError(f"Backend '{output}' not supported")

    if hasattr(main, 'fpga'):
        main.fpga.constraints(file_name);



"""
Defines a subtype of m.Array called m.Bits

m.Bits[N] is roughly equivalent ot m.Array[N, T]
"""
import weakref
from hwtypes import BitVector, SIntVector

from .compatibility import IntegerTypes
from .ref import AnonRef
from .bit import Bit, VCC, GND
from .array import ArrayType, ArrayKind
from .debug import debug_wire


__all__ = ['Bits', 'BitsType', 'BitsKind']
__all__ += ['UInt', 'UIntType', 'UIntKind']
__all__ += ['SInt', 'SIntType', 'SIntKind']
__all__ += ['BFloat', 'BFloatKind']


class BitsKind(ArrayKind):
    _class_cache = weakref.WeakValueDictionary()

    def __str__(cls):
        if not hasattr(cls, "T"):
            return cls.__name__
        if cls.isinput():
            return "In(Bits[{}])".format(cls.N)
        if cls.isoutput():
            return "Out(Bits[{}])".format(cls.N)
        return "Bits[{}]".format(cls.N)


    def __getitem__(cls, index):
        if isinstance(index, tuple):
            width, T = index
        else:
            width = index
            T = Bit
        if isinstance(width, Bits):
            assert width.const()
            # TODO: Move this logic to a method in BitsType
            bit_type_to_constant_map = {
                GND: 0,
                VCC: 1
            }
            width = BitVector[len(width)]([bit_type_to_constant_map[x] for x in
                                           width]).as_uint()
        try:
            result = BitsKind._class_cache[width, T]
            return result
        except KeyError:
            pass
        bases = [cls]
        bases = tuple(bases)
        class_name = '{}[{}, {}]'.format(cls.__name__, width, T.__name__)
        t = type(cls)(class_name, bases, dict(T=T, N=width))
        t.__module__ = cls.__module__
        BitsKind._class_cache[width, T] = t
        return t

    def qualify(cls, direction):
        if cls.T.isoriented(direction):
            return cls
        return Bits[cls.N, cls.T.qualify(direction)]

    def flip(cls):
        return Bits[cls.N, cls.T.flip()]

#     def __call__(cls, value=None, *args, **kwargs):
#         print(cls, cls.__dict__)
#         if value is not None:
#             if isinstance(value, (bool, IntegerTypes)):
#                 from .conversions import bits
#                 return bits(value, cls.N)
#             else:
#                 raise ValueError("Bit can only be initialized with None, bool, "
#                                  "or integer")
#         return super().__call__(*args, **kwargs)

    def get_family(self):
        import magma as m
        return m.get_family()


class Bits(ArrayType, metaclass=BitsKind):
    def __repr__(self):
        if not isinstance(self.name, AnonRef):
            return repr(self.name)
        ts = [repr(t) for t in self.ts]
        return 'bits([{}])'.format(', '.join(ts))

    def bits(self):
        if not self.const():
            raise Exception("Not a constant")

        def convert(x):
            if x is VCC:
                return True
            assert x is GND
            return False
        return [convert(x) for x in self.ts]

    def __int__(self):
        if not self.const():
            raise Exception("Can't call __int__ on a non-constant")
        return BitVector[len(self)](self.bits()).as_int()

    @debug_wire
    def wire(i, o, debug_info):
        if isinstance(o, IntegerTypes):
            if o.bit_length() > len(i):
                raise ValueError(
                    f"Cannot convert integer {o} (bit_length={o.bit_length()}) to Bits({len(i)})")
            from .conversions import bits
            o = bits(o, len(i))
        super().wire(o, debug_info)

    def zext(self, value):
        from .conversions import zext
        t = zext(self, value)
        return t

    def __getitem__(self, key):
        from .conversions import bits
        result = super().__getitem__(key)
        if isinstance(key, slice):
            return bits(result)
        return result


BitsType = Bits


# def Bits(N, T=None):
#     if T is None:
#         T = Bit
#     assert isinstance(N, (IntegerTypes, BitsType)), (N, type(N))
#     if isinstance(N, BitsType):
#         assert N.const()
#         # TODO: Move this logic to a method in BitsType
#         bit_type_to_constant_map = {
#             GND: 0,
#             VCC: 1
#         }
#         N = BitVector([bit_type_to_constant_map[x] for x in N]).as_uint()
#     name = 'Bits({})'.format(N)
#     return BitsKind(name, (BitsType,), dict(N=N, T=T))


class UIntKind(BitsKind):
    _class_cache = weakref.WeakValueDictionary()

    def __getitem__(cls, index):
        if isinstance(index, tuple):
            width, T = index
        else:
            width = index
            T = Bit
        if isinstance(width, UInt):
            assert width.const()
            # TODO: Move this logic to a method in BitsType
            bit_type_to_constant_map = {
                GND: 0,
                VCC: 1
            }
            width = BitVector[len(width)]([bit_type_to_constant_map[x] for x in
                                           width]).as_uint()
        try:
            return UIntKind._class_cache[width, T]
        except KeyError:
            pass
        bases = [cls]
        bases = tuple(bases)
        class_name = '{}[{}, {}]'.format(cls.__name__, width, T)
        t = type(cls)(class_name, bases, dict(T=T, N=width))
        t.__module__ = cls.__module__
        UIntKind._class_cache[width, T] = t
        return t

    def __str__(cls):
        if cls.isinput():
            return "In(UInt[{}])".format(cls.N)
        if cls.isoutput():
            return "Out(UInt[{}])".format(cls.N)
        return "UInt[{}]".format(cls.N)

    def qualify(cls, direction):
        if cls.T.isoriented(direction):
            return cls
        return UInt[cls.N, cls.T.qualify(direction)]

    def flip(cls):
        return UInt[cls.N, cls.T.flip()]


class UInt(Bits, metaclass=UIntKind):
    def __repr__(self):
        if not isinstance(self.name, AnonRef):
            return repr(self.name)
        ts = [repr(t) for t in self.ts]
        return 'uint([{}])'.format(', '.join(ts))

    def __getitem__(self, key):
        from .conversions import uint
        result = super().__getitem__(key)
        if isinstance(key, slice):
            return uint(result)
        return result


# def UInt(N, T=None):
#     if T is None:
#         T = Bit
#     assert isinstance(N, IntegerTypes)
#     name = 'UInt({})'.format(N)
#     return UIntKind(name, (UIntType,), dict(N=N, T=T))


class SIntKind(BitsKind):
    _class_cache = weakref.WeakValueDictionary()
    def __getitem__(cls, index):
        if isinstance(index, tuple):
            width, T = index
        else:
            width = index
            T = Bit
        if isinstance(width, SInt):
            assert width.const()
            # TODO: Move this logic to a method in BitsType
            bit_type_to_constant_map = {
                GND: 0,
                VCC: 1
            }
            width = BitVector[len(width)]([bit_type_to_constant_map[x] for x in
                                           width]).as_sint()
        try:
            return SIntKind._class_cache[width, T]
        except KeyError:
            pass
        bases = [cls]
        bases = tuple(bases)
        class_name = '{}[{}, {}]'.format(cls.__name__, width, T)
        t = type(cls)(class_name, bases, dict(T=T, N=width))
        t.__module__ = cls.__module__
        SIntKind._class_cache[width, T] = t
        return t

    def __str__(cls):
        if cls.isinput():
            return "In(SInt[{}])".format(cls.N)
        if cls.isoutput():
            return "Out(SInt[{}])".format(cls.N)
        return "SInt[{}]".format(cls.N)

    def qualify(cls, direction):
        if cls.T.isoriented(direction):
            return cls
        return SInt[cls.N, cls.T.qualify(direction)]

    def flip(cls):
        return SInt[cls.N, cls.T.flip()]


class SInt(Bits, metaclass=SIntKind):
    def __repr__(self):
        if not isinstance(self.name, AnonRef):
            return repr(self.name)
        ts = [repr(t) for t in self.ts]
        return 'sint([{}])'.format(', '.join(ts))

    def __int__(self):
        if not self.const():
            raise Exception("Can't call __int__ on a non-constant")
        return SIntVector[len(self)](self.bits()).as_sint()

    def sext(self, value):
        from .conversions import sext
        return sext(self, value)

UIntType = UInt
SIntType = SInt


# def SInt(N, T=None):
#     if T is None:
#         T = Bit
#     assert isinstance(N, IntegerTypes)
#     name = 'SInt({})'.format(N)
#     return SIntKind(name, (SIntType,), dict(N=N, T=T))


class BFloatKind(BitsKind):
    _class_cache = weakref.WeakValueDictionary()

    def __getitem__(cls, index):
        if isinstance(index, tuple):
            width, T = index
        else:
            width = index
            T = Bit
        if isinstance(width, BFloat):
            assert width.const()
            # TODO: Move this logic to a method in BitsType
            bit_type_to_constant_map = {
                GND: 0,
                VCC: 1
            }
            width = BitVector[len(width)]([bit_type_to_constant_map[x] for x in
                                           width]).as_uint()
        try:
            return BFloatKind._class_cache[width, T]
        except KeyError:
            pass
        bases = [cls]
        bases = tuple(bases)
        class_name = '{}[{}, {}]'.format(cls.__name__, width, T)
        t = type(cls)(class_name, bases, dict(T=T, N=width))
        t.__module__ = cls.__module__
        BFloatKind._class_cache[width, T] = t
        return t

    def __str__(cls):
        if cls.isinput():
            return "In(BFloat[{}])".format(cls.N)
        if cls.isoutput():
            return "Out(BFloat[{}])".format(cls.N)
        return "BFloat[{}]".format(cls.N)

    def qualify(cls, direction):
        if cls.T.isoriented(direction):
            return cls
        return BFloat[cls.N, cls.T.qualify(direction)]

    def flip(cls):
        return BFloat[cls.N, cls.T.flip()]


class BFloat(Bits, metaclass=BFloatKind):
    def __repr__(self):
        if not isinstance(self.name, AnonRef):
            return repr(self.name)
        ts = [repr(t) for t in self.ts]
        return 'bfloat([{}])'.format(', '.join(ts))

    def __getitem__(self, key):
        from .conversions import bfloat
        result = super().__getitem__(key)
        if isinstance(key, slice):
            return bfloat(result)
        return result



from .passes import DefinitionPass
from .clock import WireClockPass

class IRPass(DefinitionPass):
    def __init__(self, main):
        super(IRPass, self).__init__(main)
        self.code = ''

        WireClockPass(main).run()

    def __call__(self, definition):
        self.code += repr(definition) + '\n\n'



from ..is_definition import isdefinition
from ..clock import wiredefaultclock
from .passes import DefinitionPass

__all__ = ['WireClockPass']

class WireClockPass(DefinitionPass):
    def __call__(self, definition):
        for instance in definition.instances:
            wiredefaultclock(definition, instance)




import gc
from .passes import DefinitionPass

__all__ = ['DebugNamePass']

class InstDecl:
    def __init__(self, varname, lineno, filename):
        self.varname = varname
        self.lineno = lineno
        self.filename = filename

    def __repr__(self):
        return self.filename + " line " + str(self.lineno) + ": " + self.varname

def create_debugname(ref_name, ref, inst):
    if ref is inst:
        return ref_name

    if isinstance(ref, list):
        idx = ref.index(inst)
        return "{}[{}]".format(ref_name, idx)

    return ref_name

class DebugNamePass(DefinitionPass):
    def __call__(self, definition):
        for inst in definition.instances:
            referrers = gc.get_referrers(inst)

            if hasattr(inst, 'decl'):
                # Skip if this pass has already been run (partially or not)
                continue
            stack = inst.stack
            inst.decl = None
            for frame_info in stack[1:]: # Skip the first element in the stack, where the inst is placed
                local_vars = frame_info[0].f_locals.items()
                for name, var in local_vars:
                    if any(var is x for x in referrers) or var is inst:
                        debugname = create_debugname(name, var, inst)

                        # Ignore the 'self' frame (where the inst is created)
                        if debugname == 'self':
                            break

                        inst.decl = InstDecl(debugname, frame_info[2], frame_info[0].f_code.co_filename)
                        break



def tsort(graph):
    """
    Repeatedly go through all of the nodes in the graph, moving each of
    the nodes that has all its edges resolved, onto a sequence that
    forms our sorted graph. A node has all of its edges resolved and
    can be moved once all the nodes its edges point to, have been moved
    from the unsorted graph onto the sorted one.
    """

    # This is the list we'll return, that stores each node/edges pair
    # in topological order.
    graph_sorted = []

    # Convert the unsorted graph into a hash table. This gives us
    # constant-time lookup for checking if edges are unresolved, and
    # for removing nodes from the unsorted graph.
    graph_unsorted = dict(graph)

    # Run until the unsorted graph is empty.
    while graph_unsorted:

        # Go through each of the node/edges pairs in the unsorted
        # graph. If a set of edges doesn't contain any nodes that
        # haven't been resolved, that is, that are still in the
        # unsorted graph, remove the pair from the unsorted graph,
        # and append it to the sorted graph. Note here that by using
        # using the items() method for iterating, a copy of the
        # unsorted graph is used, allowing us to modify the unsorted
        # graph as we move through it. We also keep a flag for
        # checking that that graph is acyclic, which is true if any
        # nodes are resolved during each pass through the graph. If
        # not, we need to bail out as the graph therefore can't be
        # sorted.
        acyclic = False
        for node, edges in list(graph_unsorted.items()):
            for edge in edges:
                if edge in graph_unsorted:
                    break
            else:
                acyclic = True
                del graph_unsorted[node]
                graph_sorted.append((node, edges))

        if not acyclic:
            # Uh oh, we've passed through all the unsorted nodes and
            # weren't able to resolve any of them, which means there
            # are nodes with cyclic edges that will never be resolved,
            # so we bail out with an error.
            raise RuntimeError("A cyclic dependency occurred")

    return graph_sorted

if __file__ == '__main__':
    from pprint import pprint
    graph = [(2, []),
             (5, [11]),
             (11, [2, 9, 10]),
             (7, [11, 8]),
             (9, []),
             (10, []),
             (8, [9]),
             (3, [10, 8])]

    pprint(tsort(graph))
    #[(2, []),
    # (9, []),
    # (10, []),
    # (11, [2, 9, 10]),
    # (5, [11]),
    # (8, [9]),
    # (3, [10, 8]),
    # (7, [11, 8])]



from .config import get_debug_mode
from .logging import error, warning, get_source_line
from .backend.util import make_relative


__all__  = ['INPUT', 'OUTPUT', 'INOUT']
__all__ += ['flip']
__all__ += ['Port']


INPUT = 'input'
OUTPUT = 'output'
INOUT = 'inout'


def _report_wiring_messgae(fn, message, debug_info):
    if not debug_info:
        error(message)
        return
    file = debug_info[0]
    line = debug_info[1]
    message = f"\033[1m{make_relative(file)}:{line}: {message}"
    fn(message, include_wire_traceback=True)
    try:
        fn(get_source_line(file, line))
    except FileNotFoundError:
        fn(f"    Could not file file {file}")


def report_wiring_error(message, debug_info):
    _report_wiring_messgae(error, message, debug_info)


def report_wiring_warning(message, debug_info):
    # TODO(rsetaluri): Include wire traceback support.
    _report_wiring_messgae(warning, message, debug_info)


def flip(direction):
    assert direction in [INPUT, OUTPUT, INOUT]
    if   direction == INPUT:  return OUTPUT
    elif direction == OUTPUT: return INPUT
    elif direction == INOUT:  return INOUT


def merge_wires(new, old, debug_info):
    oldinputs = set(old.inputs)
    newinputs = set(new.inputs)
    oldoutputs = set(old.outputs)
    newoutputs = set(new.outputs)

    for i in oldinputs - newinputs:
        new.inputs.append(i)
        i.wires = new

    for o in oldoutputs - newoutputs:
        if len(new.outputs) > 0:
            outputs = [o.bit.debug_name for o in new.outputs]
            report_wiring_error(f"Connecting more than one output ({outputs}) to "
                                f"an input `{i.bit.debug_name}`", debug_info)
        new.outputs.append(o)
        o.wires = new


def fast_merge_wires(w, i, o):
    w.inputs = i.wires.inputs + o.wires.inputs
    w.outputs = i.wires.outputs + o.wires.outputs
    w.inputs = list(set(w.inputs))
    w.outputs = list(set(w.outputs))
    if len(w.outputs) > 1:
        outputs = [o.bit.debug_name for o in w.outputs]
        # Use w.inputs[0] as i, similar to {i.bit.debug_name}.
        report_wiring_error(f"Connecting more than one output ({outputs}) to "
                            f"an input `{w.inputs[0].bit.debug_name}`",
                            debug_info)
    for p in w.inputs:
        p.wires = w
    for p in w.outputs:
        p.wires = w


class Wire:
    """
    A Wire has a list of input and output Ports.
    """
    def __init__(self):
        self.inputs = []
        self.outputs = []

    def connect( self, o, i , debug_info):
        """
        Anon Ports are added to the input or output list of this wire.

        Connecting to a non-anonymous port to an anonymous port add the
        non-anonymous port to the wire associated with the anonymous port.
        """

        if not o.anon():
            if o.bit.isinput():
                report_wiring_error(f"Using `{o.bit.debug_name}` (an input) as "
                                    f"an output", debug_info)
                return

            if o not in self.outputs:
                if len(self.outputs) != 0:
                    output_str = ", ".join([output.bit.debug_name \
                                            for output in self.outputs])
                    msg = (f"Adding the output `{o.bit.debug_name}` to the "
                           f"wire `{i.bit.debug_name}` which already has "
                           f"output(s) `[{output_str}]`")
                    report_wiring_warning(msg, debug_info)
                self.outputs.append(o)

        if not i.anon():
            if i.bit.isoutput():
                report_wiring_error(f"Using `{i.bit.debug_name}` (an output) "
                                    f"as an input", debug_info)
                return

            if i not in self.inputs:
                self.inputs.append(i)

        # Always update wires.
        o.wires = self
        i.wires = self

    def check(self):
        for o in self.inputs:
            if o.isoutput():
                error("Output in the wire inputs: {}".format(o))

        for o in self.outputs:
            if o.isinput():
                error("Input in the wire outputs: {}".format(o))
                return False

        # Check that this wire is only driven by a single output.
        if len(self.outputs) > 1:
            error("Multiple outputs on a wire: {}".format(self.outputs))
            return False

        return True


class Port:
    """
    Ports implement wiring.

    Each port is represented by a Bit().
    """
    def __init__(self, bit):
        self.bit = bit
        self.wires = Wire()

    def __repr__(self):
        return repr(self.bit)

    def __str__(self):
        return str(self.bit)

    def anon(self):
        return self.bit.anon()

    def wire(i, o, debug_info):
        """
        Wire a port to a port.
        """
        if i.wires and o.wires and i.wires is not o.wires:
            w = Wire()
            if get_debug_mode():
                merge_wires(w, i.wires, debug_info)
                merge_wires(w, o.wires, debug_info)
            else:
                fast_merge_wires(w, i, o)
        elif o.wires:
            w = o.wires
        elif i.wires:
            w = i.wires
        else:
            w = Wire()

        w.connect(o, i, debug_info)

    def trace(self):
        """
        If the port is an input or inout, return the output.
        If the port is an output, return the first input.
        """
        if not self.wires:
            return None

        if self in self.wires.inputs:
            if len(self.wires.outputs) < 1:
                return None
            assert len(self.wires.outputs) == 1
            return self.wires.outputs[0]

        if self in self.wires.outputs:
            if len(self.wires.inputs) < 1:
                return None
            assert len(self.wires.inputs) == 1
            return self.wires.inputs[0]

        return None

    def value(self):
        """
        If the port is in the inputs, return the output.
        """
        if not self.wires:
            return None

        if self in self.wires.inputs:
            if len(self.wires.outputs) < 1:
                return None
            return self.wires.outputs[0]

        return None

    def driven(self):
        return self.value() is not None

    def wired(self):
        return self.trace() is not None



from ..is_definition import isdefinition
from .tsort import tsort

__all__ = ['Pass', 'InstancePass', 'DefinitionPass', 'InstanceGraphPass']

# abstract base class
class Pass(object):
    def __init__(self, main):
        self.main = main

    def run(self):
        self.done()
        return self

    def done(self):
        pass


class InstancePass(Pass):
    def __init__(self, main):
        super(InstancePass, self).__init__(main)
        self.instances = []

    def _run(self, definition, path):
        for instance in definition.instances:
            instancedefinition = type(instance)
            instpath = path + (instance, )
            self.instances.append(instpath)
            if callable(self):
                self(instpath)
            if isdefinition(instancedefinition):
                self._run( instancedefinition, instpath )
    
    def run(self):
         self._run(self.main, tuple())
         self.done()
         return self

class DefinitionPass(Pass):
    def __init__(self,main):
        super(DefinitionPass, self).__init__(main)
        self.definitions = {}

    def _run(self, definition):
        for instance in definition.instances:
            instancedefinition = type(instance)
            if isdefinition(instancedefinition):
                self._run( instancedefinition )

        # call each definition only once
        id_ = id(definition)
        if id_ not in self.definitions:
            self.definitions[id_] = definition
            if callable(self):
                self(definition)

    def run(self):
         self._run(self.main)
         self.done()
         return self


class BuildInstanceGraphPass(DefinitionPass):
    def __init__(self, main):
        super(BuildInstanceGraphPass, self).__init__(main)
        self.graph = {}

    def __call__(self, definition):
        if definition not in self.graph:
            self.graph[definition]  = []
        for instance in definition.instances:
            instancedefinition = type(instance)
            if instancedefinition not in self.graph:
                self.graph[instancedefinition]  = []
            if instancedefinition not in self.graph[definition]:
                #print('Adding',definition.name, instancedefinition.name)
                self.graph[definition].append(instancedefinition)

    def done(self):
        graph = []
        for vert, edges in self.graph.items():
            graph.append((vert, edges))
        self.tsortedgraph = tsort(graph)


class InstanceGraphPass(Pass):
    def __init__(self, main):
        super(InstanceGraphPass, self).__init__(main)

        p = BuildInstanceGraphPass(main).run()
        self.tsortedgraph = p.tsortedgraph

        if callable(self):
            for vert, edges in self.tsortedgraph:
                self(vert, edges)




from collections.abc import Sequence
from .bitutils import int2seq
from .ref import AnonRef, ArrayRef
from .t import Type, Kind
from .compatibility import IntegerTypes
from .bit import VCC, GND
from .bitutils import seq2int
from .debug import debug_wire, get_callee_frame_info
from .port import report_wiring_error
import weakref

__all__ = ['ArrayType', 'ArrayKind', 'Array']


class ArrayKind(Kind):
    _class_cache = weakref.WeakValueDictionary()
    def __init__(cls, name, bases, dct):
        Kind.__init__(cls, name, bases, dct)

    def __str__(cls):
        return "Array[%d, %s]" % (cls.N, cls.T)

    def __getitem__(cls, index):
        width, sub_type = index
        try:
            return ArrayKind._class_cache[width, sub_type]
        except KeyError:
            pass
        bases = [cls]
        bases = tuple(bases)
        class_name = '{}[{}, {}]'.format(cls.__name__, width, sub_type.__name__)
        t = type(cls)(class_name, bases, dict(N=width, T=sub_type))
        t.__module__ = cls.__module__
        ArrayKind._class_cache[width, sub_type] = t
        return t

    def __eq__(cls, rhs):
        if not issubclass(type(rhs), ArrayKind):
            return False

        if cls.N != rhs.N:
            return False
        if cls.T != rhs.T:
            return False

        return True

    __ne__ = Kind.__ne__
    __hash__ = Kind.__hash__

    def __len__(cls):
        return cls.N

#     def __getitem__(cls, key):
#         if isinstance(key, slice):
#             return array([cls[i] for i in range(*key.indices(len(cls)))])
#         else:
#             if not (0 <= key and key < cls.N):
#                 raise IndexError

#             return cls.ts[key]

    def size(cls):
        return cls.N * cls.T.size()

    def qualify(cls, direction):
        if cls.T.isoriented(direction):
            return cls
        return Array[cls.N, cls.T.qualify(direction)]

    def flip(cls):
        return Array[cls.N, cls.T.flip()]

    def __call__(cls, *args, **kwargs):
        result = super().__call__(*args, **kwargs)
        if len(args) == 1 and isinstance(args[0], Array) and not \
                (issubclass(cls.T, Array) and cls.N == 1):
            arg = args[0]
            if len(arg) < len(result):
                from .conversions import zext
                arg = zext(arg, len(result) - len(arg))
            result(arg)
        return result


class Array(Type, metaclass=ArrayKind):
    def __init__(self, *largs, **kwargs):

        Type.__init__(self, **kwargs)

        if isinstance(largs, Sequence) and len(largs) == self.N:
            self.ts = []
            for t in largs:
                if isinstance(t, IntegerTypes):
                    t = VCC if t else GND
                assert type(t) == self.T or isinstance(type(t), type(self.T)), \
                    (type(t), self.T)
                self.ts.append(t)
        elif len(largs) == 1 and isinstance(largs[0], int):
            self.ts = []
            for bit in int2seq(largs[0], self.N):
                self.ts.append(VCC if bit else GND)
        elif len(largs) == 1 and isinstance(largs[0], Array):
            assert len(largs[0]) <= self.N
            T = self.T
            self.ts = list(type(t)() for t in largs[0])
            if len(largs[0]) < self.N:
                self.ts += [self.T() for _ in range(self.N - len(largs[0]))]
        elif len(largs) == 1 and isinstance(largs[0], list):
            assert len(largs[0]) <= self.N
            T = self.T
            self.ts = largs[0][:]
            if len(largs[0]) < self.N:
                self.ts += [self.T() for _ in range(self.N - len(largs[0]))]
        else:
            self.ts = []
            for i in range(self.N):
                T = self.T
                t = T(name=ArrayRef(self, i))
                self.ts.append(t)

    def __eq__(self, rhs):
        if not isinstance(rhs, ArrayType):
            return False
        return self.ts == rhs.ts

    __hash__ = Type.__hash__

    def __repr__(self):
        if not isinstance(self.name, AnonRef):
            return repr(self.name)
        ts = [repr(t) for t in self.ts]
        return 'array([{}])'.format(', '.join(ts))

    def __len__(self):
        return self.N

    def __getitem__(self, key):
        if isinstance(key, ArrayType) and all(t in {VCC, GND} for t in key.ts):
            key = seq2int([0 if t is GND else 1 for t in key.ts])
        if isinstance(key, slice):
            return array([self[i] for i in range(*key.indices(len(self)))])
        else:
            if not (-self.N <= key and key < self.N):
                raise IndexError

            return self.ts[key]

    def __add__(self, other):
        other_len = other.N
        total = self.N + other_len
        res_bits = []
        for i in range(total):
            res_bits.append(self[i] if i < self.N else other[i - self.N])
        return array(res_bits)

    def __call__(self, o):
        return self.wire(o, get_callee_frame_info())

    @classmethod
    def isoriented(cls, direction):
        return cls.T.isoriented(direction)

    def as_list(self):
        return [self[i] for i in range(len(self))]

    @debug_wire
    def wire(i, o, debug_info):
        # print('Array.wire(', o, ', ', i, ')')

        if not isinstance(o, ArrayType):
            if isinstance(o, IntegerTypes):
                report_wiring_error(f'Cannot wire {o} (type={type(o)}) to {i.debug_name} (type={type(i)}) because conversions from IntegerTypes are only defined for Bits, not general Arrays', debug_info)  # noqa
            else:
                report_wiring_error(f'Cannot wire {o.debug_name} (type={type(o)}) to {i.debug_name} (type={type(i)}) because {o.debug_name} is not an Array', debug_info)  # noqa
            return

        if i.N != o.N:
            report_wiring_error(f'Cannot wire {o.debug_name} (type={type(o)}) to {i.debug_name} (type={type(i)}) because the arrays do not have the same length', debug_info)  # noqa
            return

        for k in range(len(i)):
            i[k].wire(o[k], debug_info)

    def driven(self):
        for t in self.ts:
            if not t.driven():
                return False
        return True

    def wired(self):
        for t in self.ts:
            if not t.wired():
                return False
        return True

    # test whether the values refer a whole array
    def iswhole(self, ts):

        n = len(ts)

        for i in range(n):
            if ts[i].anon():
                return False

        for i in range(n):
            # elements must be an array reference
            if not isinstance(ts[i].name, ArrayRef):
                return False

        for i in range(1, n):
            # elements must refer to the same array
            if ts[i].name.array is not ts[i-1].name.array:
                return False

        if n > 0 and n != ts[0].name.array.N:
            # must use all of the elements of the base array
            return False

        for i in range(n):
            # elements should be numbered consecutively
            if ts[i].name.index != i:
                return False

        return True

    def trace(self):
        ts = [t.trace() for t in self.ts]

        for t in ts:
            if t is None:
                return None

        if self.iswhole(ts):
            return ts[0].name.array

        return array(ts)

    def value(self):
        ts = [t.value() for t in self.ts]

        for t in ts:
            if t is None:
                return None

        if self.iswhole(ts):
            return ts[0].name.array

        return array(ts)

    def const(self):
        for t in self.ts:
            if not t.const():
                return False

        return True

    def flatten(self):
        return sum([t.flatten() for t in self.ts], [])

    def concat(self, *args):
        return concat(self, *args)


# def Array(N, T):
#     assert isinstance(N, IntegerTypes)
#     assert isinstance(T, Kind)
#     name = 'Array(%d,%s)' % (N, str(T))
#     return ArrayKind(name, (ArrayType,), dict(N=N, T=T))

ArrayType = Array


# Workaround for circular dependency
from .conversions import array, concat  # nopep8
